##
# File:    metaDataobject.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#
##

"""
holds the data information of basic workflow class
"""

__docformat__ = "restructuredtext en"
__author__ = "Tom Oldfield"
__email__ = "oldfield@ebi.ac.uk"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.24"


class metaDataObject:
    def __init__(self):

        self.ID = None
        self.versionMajor = None
        self.versionMinor = None
        self.versionAuthor = None
        self.name = None
        self.date = None
        self.description = None
        self.author = None

    def setID(self, did):
        self.ID = did

    def setVersionMajor(self, version):
        self.versionMajor = version

    def setVersionMinor(self, version):
        self.versionMinor = version

    def setDate(self, name):
        self.date = name

    def setAuthor(self, name):
        self.author = name

    def setName(self, name):
        self.name = name

    def setDescription(self, name):
        self.description = name

    def getID(self):
        return self.ID

    def getVersionMajor(self):
        return self.versionMajor

    def getVersionMinor(self):
        return self.versionMinor

    def getAuthor(self):
        return self.author

    def getName(self):
        return self.name

    def getDescription(self):
        return self.description



##
# File: ServerMonitor.py
# Date: ~ 2009
#
#  Updates:
#     10-Mar-2015   jdw refactored
#     25-Apr-2015   jdw purge unused code sections
#     8-May-2015    jdw reorganized sleep/wake processing - handle exit conditions
#     3-Jun-2016    jdw Overall watch method to resolve restart and session import issues
#
##

import time
import signal
import socket
import logging


from wwpdb.apps.wf_engine.engine.WorkflowManager import WorkflowManager

from wwpdb.apps.wf_engine.wf_engine_utils.run.MyLogger import MyLogger
from wwpdb.apps.wf_engine.wf_engine_utils.process.ProcessUtils import ProcessUtils
from wwpdb.apps.wf_engine.engine.EngineUtils import EngineUtils
from wwpdb.apps.wf_engine.wf_engine_utils.time.TimeStamp import TimeStamp

logger = logging.getLogger(name="root")


class ServerMonitor(object):

    """

     Original author T.J.Oldfield 2009


     Main server  demon
      manages processes by pinging statusDB

    Checks for processes that need to be run
    checks for kills/restarts for processes it owns
    checks on other server demons to see if they are broken
    checks and puts to sleep tired processes
    checks and wakes up processes

    puts its current state into the statusDB (load and onership)
    """

    def __init__(self, instID, classID, depID, debug=0):  # pylint: disable=unused-argument
        self.__baseEngineProcessName = "WFEngineRunnerExec"
        self.rowLim = 1
        self.wait = False
        self.locking = False
        #    self.inStatus = "init"
        #    self.outStatus = "init"
        self.inStatus = "runWF"
        self.outStatus = "runningWF"
        #  watch loop delay interval - seconds
        self.waitTime = 5
        # not used
        # self.sleepTime = 3600
        self.classID = classID
        self.depID = depID
        self.monitorDict = {}
        self.extremeLoad = 16
        self.normalLoad = 8
        # ~ idle timeout value - seconds
        # self.purgeTime = 60
        self.purgeTime = 600

        self.debug = debug
        self.__lfh = MyLogger(level=logging.INFO)

        self.hostname = socket.gethostname()

        ##
        # JDW  - creating new connection here -- Can this be passed along ??
        self.__eUtil = EngineUtils(verbose=self.debug)
        self.__timeStamp = TimeStamp()

        self.validSleep = ["waiting", "open"]

        self.__pu = ProcessUtils(verbose=True, log=self.__lfh)
        # Temporary setting to track process monitoring
        # self.__pu.setDebug(flag=True)
        logger.info("+ServerMonitor.__init__() Normal load %s and high load is %s ", self.normalLoad, self.extremeLoad)

    def watch(self):
        """
        Montitoring method that looks at the communication table in the database for new tasks.

        JDW - This is the ONLY public entry point in this class -
        """

        logger.info("+ServerMonitor.watch() Starting watch process")

        # first - what does this server own - get state of communication table
        # take this out as the status module does not respect concurrency and so the restart
        #  overwrites any mmCIF changes on recover
        #    self.__recoverState()

        # use this skip work
        newWork = 1
        numMonitored = 0
        while True:
            for loop in range(1, 10000000):
                #      do a reconnect after timeout - stop stale conenctions
                self.__eUtil.updateConnection()

                if loop == 1:
                    self.__monitorRestart(verbose=True)
                #
                dictList = []
                if (loop % newWork) == 0:
                    # Look for the next PENDING workflow, add to the monitored list, and invoke.
                    dictList = self.__getStatusData(self.inStatus, self.outStatus)
                    if dictList is not None and len(dictList) > 0:
                        # start new workflows
                        self.__addProcess(dictList)

                # manage monitored PENDING entries in communications table belonging to this server
                self.__manageProcess()

                time.sleep(self.waitTime)  # wait for x seconds

                # newWork will be adjusted to 1/2/3 depending on the number of managed processes --
                newWork = self.__getWorkStep()

                # check the process workflows looking for WF in a wait state
                if loop % 2 == 0:
                    self.__sleepWorkFlow()

                if (loop + 1) % 2 == 0:
                    self.__wakeUpWorkflow()

                if loop % 5 == 0:
                    self.__monitorRestart(verbose=True)

                if len(self.monitorDict) != numMonitored:
                    numMonitored = len(self.monitorDict)
                    self.__showMonitored()

    def __addProcess(self, dictList):
        """
        Starts a PENDING workflow based on the tasks described in the input dictionary list.

        Tasks are also copied to the global self.monitorDict for this serverMonitor() instance

         >> sets all previous wf_instance for depID to exception if not already exception (preserve time)?
        """
        numT = len(dictList)
        for ii, edict in enumerate(dictList):
            logger.info("+ServerMonitor.__addProcess begin iteration %d or %d", ii + 1, numT)
            logger.info("+ServerMonitor.__addProcess using dictionary =  %s", edict.items())
            depID = edict["DEP_SET_ID"]
            classID = edict["WF_CLASS_ID"]
            # instID = edict['WF_INST_ID']
            wfName = edict["WF_CLASS_FILE"]
            if wfName is None:
                logger.info("+ServerMonitor.__addProcess No workflow class name found - ignoring ")
            else:
                logger.info("+ServerMonitor.__addProcess for %s classID %s wfName %s ", depID, classID, wfName)
                n = wfName.find(":")
                # look up a workflow nesting of task
                if n > 0:
                    taskID = wfName[n + 1 :]
                    wfName = wfName[:n]
                    if (taskID == "Annotate") and (wfName == "Annotation.bf.xml"):
                        taskID = "entry-point"
                else:
                    taskID = "entry-point"
                logger.info("+ServerMonitor.__addProcess for %s wfName %s using taskID %s ", depID, wfName, taskID)
                #
                # set all the old instances to exception
                #
                replaceStatus = "aborted"
                logger.info("+ServerMonitor.__addProcess %s setting active workflow instances to %s", depID, replaceStatus)
                sql = (
                    "update wf_instance set inst_status = '" + replaceStatus + "' where dep_set_id = '" + str(depID) + "' and inst_status not in ('exception','aborted','finished')"
                )
                _ok = self.__eUtil.runUpdateSQL(sql)  # noqa: F841

                logger.info("+ServerMonitor.__addProcess starting workflow manager for %s classId %s wfName %s taskID %s", depID, classID, wfName, taskID)
                wf = WorkflowManager(wfName, depID, None, taskID)
                pid = wf.runWF(wait=False)
                m = {}
                m["depID"] = depID
                m["classID"] = classID
                m["pid"] = pid
                m["timeStamp"] = self.__timeStamp.getSecondsFromReference()
                m["status"] = "working"
                self.monitorDict[depID] = m
                #
                logger.info(
                    "+ServerMonitor.__addProcess end iteration %d of %d for %s classID %s wfName %s taskId %s  status %s", ii + 1, numT, depID, classID, wfName, taskID, m["status"]
                )

        logger.info("+ServerMonitor.__addProcess leaving after %d iterations", numT)

    def __deleteProcess(self, depid, value):  # pylint: disable=unused-argument
        """
        Removes a server engine dictionary entry  value is
        """
        del value["depID"]
        del value["pid"]
        del value["classID"]
        del value["timeStamp"]
        del value["status"]

    def __createNewBucket(self, depID, classID, file, now):
        """
        method to create a new instance entry as a place holder for a WF not previously run
        """
        ts = self.__timeStamp.getTimeStringLocal(now)
        logger.info("+ServerMonitor.__createNewBucket - begins for %s classID %s file %s timeStamp %s", depID, classID, file, ts)
        try:
            wfInstId = self.__eUtil.getNextInstanceId(depID)
            #
            sql = (
                "insert into wf_instance (wf_inst_id, wf_class_id, dep_set_id, owner, inst_status, status_timestamp) values ('"
                + wfInstId
                + "','"
                + classID
                + "','"
                + depID
                + "','"
                + file
                + "','init','"
                + str(now)
                + "')"
            )

            logger.info("+ServerMonitor.__createNewBucket - for %s classID %s owner/file %s with new wfInstID %s", depID, classID, file, wfInstId)
            logger.info("+ServerMonitor.__createNewBucket  - new bucket %s", str(sql))
            ok = self.__eUtil.runUpdateSQL(sql)
            if ok == 1:
                logger.info("+ServerMonitor.__createNewBucket wf_instance insert succeeded for %s", str(depID))
            else:
                logger.info("+ServerMonitor.__createNewBucket wf_instance insert failed for %s", str(depID))

            sql = "select dep_set_id from wf_instance_last where dep_set_id = '" + depID + "'"
            ok = self.__eUtil.runSelectSQL(sql)

            if ok is None or len(ok) < 1:
                sql = (
                    "insert into wf_instance_last (wf_inst_id, wf_class_id, dep_set_id, owner, inst_status, status_timestamp) values ('"
                    + wfInstId
                    + "','"
                    + classID
                    + "','"
                    + depID
                    + "','"
                    + file
                    + "','init','"
                    + str(now)
                    + "')"
                )
                ok = self.__eUtil.runUpdateSQL(sql)
            else:
                sql = (
                    "update wf_instance_last set wf_inst_id = '"
                    + wfInstId
                    + "', wf_class_id = '"
                    + classID
                    + "', owner = '"
                    + file
                    + "', inst_status = 'init', status_timestamp = '"
                    + str(now)
                    + "' where dep_set_id = '"
                    + depID
                    + "'"
                )
                ok = self.__eUtil.runUpdateSQL(sql)
        except Exception as e:
            logger.info("+ServerMonitor.__createNewBucket for %s function failed %s", depID, str(e))
            ok = 0

        if ok < 1:
            logger.info("+ServerMonitor.__createNewBucket failed insert in wf_instance_last for %s", str(depID))
            return False
        else:
            logger.info("+ServerMonitor.__createNewBucket wf_instance set %s classID %s owner/file %s wfInstID %s status init at %s", depID, classID, file, wfInstId, ts)
            logger.info("+ServerMonitor.__createNewBucket wf_instance_last insert succeeded for %s", str(depID))

            return True

    def __manageGetInstanceRow(self, timeNow, allRow):
        """
        Private method to get the last ordinal within the wf_instance table for a dep_id / class_id


        """

        lastOrdinal = -1

        # first - get the last wf_instance record for that ID
        #        sql = "select ordinal from wf_instance where dep_set_id = '" + str(allRow[1]) + "' and wf_class_id = '" + str(allRow[2]) + "' order by status_timestamp limit 1"
        # October-7th put in desc order :  DO NOT REMOVE CLASSID as it defines where the retart occured - which might not be the last.

        sql = (
            "select ordinal, wf_inst_id from wf_instance where dep_set_id = '"
            + str(allRow[1])
            + "' and wf_class_id = '"
            + str(allRow[2])
            + "' order by status_timestamp desc limit 1"
        )
        last = self.__eUtil.runSelectSQL(sql)
        # fail state 1 : we did not get a row - bum communication
        logger.info("+ServerMonitor.__manageGetInstanceRow - Last status from wf_instance depID %s and classId %s status %r", allRow[1], allRow[2], last)
        if last is None or len(last) < 1:
            logger.info("+ServerMonitor.__manageGetInstanceRow - %s no status returned - target row %r ", allRow[1], allRow)
            # create a bucket holder  - since WF always create a new one
            #
            if not self.__createNewBucket(allRow[1], allRow[2], allRow[6], timeNow):
                logger.info("+ServerMonitor.__manageGetInstanceRow  Failed communication depId %s %s", allRow[1], allRow[4])
                sql = "update communication set status = 'failed', actual_timestamp = " + str(timeNow) + " where ordinal = " + str(allRow[0])
                _ok = self.__eUtil.runUpdateSQL(sql)  # noqa: F841
                return -1, -1, -1
            else:
                sql = (
                    "select ordinal, wf_inst_id from wf_instance where dep_set_id = '"
                    + str(allRow[1])
                    + "' and wf_class_id = '"
                    + str(allRow[2])
                    + "' order by status_timestamp desc limit 1"
                )
                last = self.__eUtil.runSelectSQL(sql)

        for ll in last:
            lastOrdinal = ll[0]
            lastInst = ll[1]
            break

        if lastOrdinal < 0:
            logger.info("+ServerMonitor.__manageGetInstanceRow  :  Failed communication %s: %s", allRow[1], allRow[4])
            sql = "update communication set status = 'failed', actual_timestamp = " + str(timeNow) + " where ordinal = " + str(allRow[0])
            _ok = self.__eUtil.runUpdateSQL(sql)  # noqa: F841
            return -1, -1, -1

        logger.info("+ServerMonitor.__manageGetInstanceRow - classId %s lastInst %s lastOrdinal %s", allRow[2], lastInst, str(lastOrdinal))
        return lastOrdinal, lastInst, allRow[2]

    def __manageKillProcess(self, timeNow, lastOrdinal, allRow):  # pylint: disable=unused-argument
        """
        kill the process and all its children,
        This only works with processes owned by this server - and server is active (internal list)
        """
        logger.info("+ServerMonitor.__manageKillProcess starts for %s", allRow[1])
        if allRow[1] not in self.monitorDict:
            return

        value = self.monitorDict[allRow[1]]
        # check that the pid is valid - if the process is sleeping then it is NONE
        if value["pid"] is not None:
            value["status"] = "sleeping"
            # JDW -- add my process utils -
            # JDW 5-Apr-2015 add parent to child list - fix kill followed restart problem
            pidList = []
            pidList.append(value["pid"].pid)
            pidList.extend(self.__pu.getChildren(value["pid"].pid))
            logger.info("+ServerMonitor.__manageKillProcess for %s class %s pidList %r", value["depID"], value["classID"], pidList)
            self.__pu.killProcessList(pidList[::-1], mySignal=signal.SIGKILL)
        #
        logger.info("+ServerMonitor.__manageKillProcess leaving for %s", allRow[1])

    def __manageResetUpdate(self, timeNow, lastOrdinal, lastInst, lastClass, allRow):
        """
        update the reset code and change the running instance to reset
        """
        logger.info(
            "+ServerMonitor.__manageResetUpdate : starts for deposition %s lastClass %s lastInst %s lastOrdinal %r - allrow %r", allRow[1], lastClass, lastInst, lastOrdinal, allRow
        )

        sql = "update communication set status = 'INIT', activity = 'INIT',actual_timestamp = " + str(timeNow) + " where ordinal = " + str(allRow[0])
        nrow = self.__eUtil.runUpdateSQL(sql)

        depID = allRow[1]

        #  set all previous instances - take off 1 second so we don't get time overlap
        # removed 7th Nov - but back in temporarily
        # #
        sql = (
            "update wf_instance set inst_status = 'aborted', status_timestamp = "
            + str(timeNow - 1)
            + " where dep_set_id = '"
            + str(depID)
            + "' and inst_status not in ('exception','finished')"
        )
        _ok = self.__eUtil.runUpdateSQL(sql)  # noqa: F841

        # set the last instance to init
        sql = "update wf_instance set inst_status = 'init', status_timestamp = " + str(timeNow) + ", owner = '" + str(allRow[6]) + "' where ordinal = " + str(lastOrdinal)
        _ok = self.__eUtil.runUpdateSQL(sql)  # noqa: F841

        # Only need the one sql as we only have one reference for this depID
        sql = (
            "update wf_instance_last set wf_inst_id = '"
            + lastInst
            + "', inst_status = 'init', wf_class_id = '"
            + str(lastClass)
            + "', owner = 'Annotation.bf.xml', status_timestamp = "
            + str(timeNow)
            + ", owner = '"
            + str(allRow[6])
            + "' where dep_set_id = '"
            + str(depID)
            + "'"
        )
        _ok = self.__eUtil.runUpdateSQL(sql)  # noqa: F841

        logger.info("+ServerMonitor.__manageResetUpdate :  Set deposition %s  back to init (%s)\n", str(allRow[1]), str(nrow))

    def __manageWaitingUpdate(self, timeNow, lastOrdinal, allRow):
        """
        update the open code and change the running instance to waiting
        do not change the activity state
        """

        sql = "update communication set status = 'WORKING', actual_timestamp = " + str(timeNow) + " where ordinal = " + str(allRow[0])
        # OLD          sql = "update wf_instance set inst_status = 'waitingWF' where ordinal = " + str(allRow[0])
        nrow = self.__eUtil.runUpdateSQL(sql)
        # sql = "update wf_instance set inst_status = 'waiting', status_timestamp
        # = " + str(timeNow) + " where dep_set_id = '" + str(allRow[1]) + "' and
        # wf_class_id = '" + str(allRow[2]) + "' and wf_inst_id = '" +
        # str(allRow[3]) + "'"
        sql = "update wf_instance set inst_status = 'waiting', status_timestamp = " + str(timeNow) + " where ordinal = " + str(lastOrdinal)
        nrow = self.__eUtil.runUpdateSQL(sql)
        logger.info("+ServerMonitor.__manageWaitingUpdate :  Set %s open interface back to waiting (%s)\n", str(allRow[1]), str(nrow))

        sql = "update wf_instance_last set inst_status = 'waiting', status_timestamp = " + str(timeNow) + " where dep_set_id = '" + str(allRow[1]) + "'"
        nrow = self.__eUtil.runUpdateSQL(sql)

    def __manageProcess(self):
        """
        Look for PENDING status for workflows already ASSIGNED to this host (e.g. host in communication table) but not necessarily monitored by this
        server process.

        method to look for status codes in the communication that need an action
        to reset the workflow for an ID
        killWF : will kill all processes associated with that ID
                  waitWF : will unlock an open interface : setting "open" to "waiting"

        restartWF : will do a kill followed by a status code reset
        restartGoWF : will do a kill, status code reset, and reRun the WF


        WARNING :  contains work around for WFM not setting the wf_intance.
                   uses the ordinal of the last row "lastOrdinal"

        """
        #
        # JDW -- This scans for pending entries in communication table ASSIGNED to the current host -
        #
        # get all the depositions owned by this monitor from the the database
        sql = (
            "select ordinal,dep_set_id, parent_wf_class_id,parent_wf_inst_id, command, status, wf_class_file from communication where upper(status) = 'PENDING' and host = '"
            + str(self.hostname)
            + "'"
        )

        allList = self.__eUtil.runSelectSQL(sql)
        if (allList is None) or (len(allList) < 1):
            # logger.info("+ServerMonitor.__manageProcess() starting with NO pending entries")
            # jdw just return here as there is nothing to do
            return
        else:
            logger.info("+ServerMonitor.__manageProcess() starting for %d PENDING entries", len(allList))

        # Check for any missing entries in internal monitorDict

        if allList is not None:
            for row in allList:
                if row[1] in self.monitorDict.keys():
                    logger.info("+ServerMonitor.__manageProcess found monitored entry %s status %s command %s", str(row[1]), str(row[5]), str(row[4]))
                    # pass
                else:
                    #           create new dictionary item
                    m = {}
                    m["depID"] = row[1]
                    m["classID"] = row[2]
                    m["pid"] = None  # it cannot be running - this is the OBJECT not the pid value in the object !
                    m["timeStamp"] = self.__timeStamp.getSecondsFromReference()
                    m["status"] = "working"
                    self.monitorDict[row[1]] = m
                    logger.info("+ServerMonitor.__manageProcess  adding entry to the MONITORED list %s list length %s", str(row[1]), str(len(self.monitorDict)))

        timeNow = self.__timeStamp.getSecondsFromReference()
        if allList is not None:
            for allRow in allList:
                lastOrdinal, lastInst, lastClass = self.__manageGetInstanceRow(timeNow, allRow)
                logger.info("+ServerMonitor.__manageProcess - lastOrdinal %s lastInst %s lastClass %s current command %s", lastOrdinal, lastInst, lastClass, allRow[4])
                if lastOrdinal < 0:
                    return
                # just kills the process cascade
                if allRow[4] == "killWF":
                    logger.info("+ServerMonitor.__manageProcess :  killing process (killWF)\n")
                    # logger.flush()
                    self.__manageKillProcess(timeNow, lastOrdinal, allRow)
                    #          self.__manageKillUpdate(timeNow,lastOrdinal,lastInst,allRow)
                    self.__manageResetUpdate(timeNow, lastOrdinal, lastInst, lastClass, allRow)
                elif allRow[4] == "restartWF":
                    logger.info("+ServerMonitor.__manageProcess :  resetting process (restartWF)\n")
                    # This actually does not start the WF - just marks it ready to start
                    # logger.flush()
                    self.__manageKillProcess(timeNow, lastOrdinal, allRow)
                    self.__manageResetUpdate(timeNow, lastOrdinal, lastInst, lastClass, allRow)
                elif allRow[4] == "restartGoWF":
                    logger.info("+ServerMonitor.__manageProcess :  resetting process (restartGoWF)\n")
                    # This actually does the WF
                    # logger.flush()
                    self.__manageKillProcess(timeNow, lastOrdinal, allRow)
                    self.__manageResetUpdate(timeNow, lastOrdinal, lastInst, lastClass, allRow)
                    self.__markRunTest(allRow[1])

                elif allRow[4] == "waitWF":
                    logger.info("+ServerMonitor.__manageProcess :  resetting interface to waiting (waitWF)\n")
                    # logger.flush()
                    # this will set the code to waiting regardless : assumes WFM is correct
                    self.__manageWaitingUpdate(timeNow, lastOrdinal, allRow)

        logger.info("+ServerMonitor.__manageProcess() leaving after processing %d PENDING entries", len(allList))

        # logger.flush()

    def __markRunTest(self, depid):

        timeNow = self.__timeStamp.getSecondsFromReference()
        self.__markRun(timeNow, depid)

    def __markRun(self, timestamp, depid):

        sql = "update communication set command = 'runWF', actual_timestamp = " + str(timestamp) + ", receiver = 'WFE', status = 'PENDING' where dep_set_id = '" + str(depid) + "'"
        ret = self.__eUtil.runUpdateSQL(sql)

        logger.info(" ServerMonitor.__markRun update command %s ", sql)
        logger.info(" ServerMonitor.__markRun update returns %r ", ret)

        sql = "select dep_set_id, parent_wf_class_id, wf_class_id, wf_inst_id,status, activity, actual_timestamp from communication where  dep_set_id = '" + str(depid) + "'"
        myList = self.__eUtil.runSelectSQL(sql)
        logger.info("+ServerMonitor.__markRun : - myList  %r", myList)

    def __checkProcess(self):
        """
        Self monitor of the internal processs managed by this instance
        rewrite using the communication table : but instID and classID not filled
          select dep_set_id, wf_class_id, wf_inst_id,status, activity from communication
                  where dep_set_id in ('  ....... ')
        """

        ngather = 0
        nlost = 0
        nsleep = 0
        gather = ""
        ncomplete = 0
        load = 0
        delList = []

        bored = 100
        myList = None

        #
        if len(self.monitorDict) > 0:
            # JDW 5-Apr-2015 -- change to parent_wf_class_id local to this method -
            sql = "select dep_set_id, parent_wf_class_id, wf_inst_id,status, activity, actual_timestamp from communication where  dep_set_id in ("
            for depid, value in self.monitorDict.items():
                sql = sql + "'" + depid + "',"
            sql = sql[:-1] + ")"
            myList = self.__eUtil.runSelectSQL(sql)
            logger.info("+ServerMonitor.__checkProcess :  monitored status  - current monitored ID list is %r", myList)

        if myList is not None:
            for row in myList:
                depid = row[0]
                status = row[3]
                activity = row[4]
                myClass = row[1]

                if status is None or activity is None:
                    logger.debug("+ServerMonitor.__checkProcess :  Skipping %s parent class %s has status %s", str(depid), myClass, str(status))
                    continue

                if status.upper() == "WORKING" and activity.upper() == "WORKING":
                    if depid in self.monitorDict:
                        mon = self.monitorDict[depid]
                        if mon is not None and "pid" in mon:
                            if mon["pid"] is not None:
                                if mon["pid"].poll() is None:
                                    # the process has not return code - so is still running
                                    load = load + 1
                                    ngather = ngather + 1
                                    if len(gather) < 200:
                                        gather = gather + "[" + str(depid) + ",working,-]" + ";"
                                else:
                                    ncomplete = ncomplete + 1
                                    # pass
                    # return code of completed process
                    #        print " id = " , depid , "   " , instID

                elif status.upper() == "INIT":
                    # If this server owns this but nothing has happened for 1 hour - drop it
                    howlong = self.__timeStamp.getSecondsFromReference() - row[5]
                    if howlong > bored and howlong < 10000000:
                        # only delete the reference if we have been checking for 1 hour
                        # this will be picked up later from the WFM
                        logger.info("+ServerMonitor.__checkProcess %s with INIT status added to delete list", str(depid))
                        delList.append(depid)

                    nlost = nlost + 1
                elif status.upper() == "FINISHED":
                    # If this server owns this but nothing has happened for 1 hour - drop it
                    # Workflow finished correctly - no way of starting again - temporary fix
                    howlong = self.__timeStamp.getSecondsFromReference() - row[5]
                    if howlong > bored and howlong < 10000000:
                        # only delete the reference if we have been checking for 1000 seconds
                        logger.info("+ServerMonitor.__checkProcess %s with FINISHED status added to delete list", str(depid))
                        delList.append(depid)
                    nlost = nlost + 1
                elif status.upper() == "EXCEPTION" or status[-1].upper() == "X":
                    # logger.debug("+ServerMonitor.__checkProcess :  Workflow for " + str(depid) + " parent class " + myClass + " has status " + str(status))
                    # logger.flush()
                    nlost = nlost + 1
                elif status.upper() == "WORKING" and activity.upper() == "SLEEPING":
                    # actually not done as I filter for working processes
                    load = load + 0.001
                    nsleep = nsleep + 1
                    ngather = ngather + 1
                    if len(gather) < 200:
                        pass
                #               gather = gather + "[" + str(depid) + ",sleeping,-]" + ";"

        for depid in delList:
            # temporary fix - what the hell ?
            if depid in self.monitorDict:
                mon = self.monitorDict[depid]
                if self.__timeStamp.getSecondsFromReference() - mon["timeStamp"] > bored:
                    # the action in the internal monitor is old too (ie we have not just tried to do something)
                    logger.info("+ServerMonitor.__checkProcess  deleting %s from monitor dictionary - current length %s", str(depid), str(len(self.monitorDict)))
                    value = self.monitorDict[depid]
                    # jdw this does nothing
                    self.__deleteProcess(depid, value)
                    del self.monitorDict[depid]

        logger.debug("+ServerMonitor.__checkProcess  host %s load %d nsleep %d nlost %d ngather %d gather %r ", self.hostname, load, nsleep, nlost, ngather, gather)
        self.__updateResourceStatus(gather)

        return load

    def __updateResourceStatus(self, gather):
        """
        Process to monitor the server load - writes a status entry

        # JDW -- not sure how this is used.  Fixed all of the backend calls --

        """
        logger.debug("+ServerMonitor.__updateResourceStatus() - starting host %s engine_monitoring update ", self.hostname)
        if len(gather) > 3:
            if gather.endswith(";"):
                gather = gather[:-1]
        #
        MemInfo = self.__pu.getMemoryInfo()
        CpuInfo = self.__pu.getCpuInfo()

        sql = "select hostname from engine_monitoring where hostname = '" + self.hostname + "'"
        myList = self.__eUtil.runSelectSQL(sql)

        timeNow = self.__timeStamp.getSecondsFromReference()
        #
        if len(myList) > 0:
            sql = (
                "update engine_monitoring set  total_physical_mem = "
                + str(MemInfo["MemTotal"])
                + ", total_virtual_mem = "
                + str(MemInfo["TotalTotal"])
                + ", physical_mem_usage = "
                + str(MemInfo["MemUsed"])
                + ", virtual_mem_usage = "
                + str(MemInfo["SwapUsed"])
                + ", swap_total = "
                + str(MemInfo["SwapTotal"])
                + ", swap_used = "
                + str(MemInfo["SwapUsed"])
                + ", swap_free = "
                + str(MemInfo["SwapFree"])
                + ", cached = "
                + str(MemInfo["Cached"])
                + ", buffers = "
                + str(MemInfo["Buffers"])
                + ",cpu_number = "
                + str(CpuInfo["Ncpu"])
                + " , cpu_usage = "
                + str(CpuInfo["Usage"])
                + ", ids_set = '"
                + gather
                + "' , status_timestamp = "
                + str(timeNow)
                + " where hostname = '"
                + str(self.hostname)
                + "'"
            )
        else:
            sql = (
                "insert engine_monitoring (hostname,total_physical_mem,total_virtual_mem,physical_mem_usage,virtual_mem_usage,swap_total,swap_used,swap_free,cached,buffers,cpu_number,cpu_usage,ids_set,status_timestamp) values ('"  # noqa: E501
                + str(self.hostname)
                + "', "
                + str(MemInfo["MemTotal"])
                + ","
                + str(MemInfo["TotalTotal"])
                + " , "
                + str(MemInfo["MemUsed"])
                + ", "
                + str(MemInfo["SwapUsed"])
                + ", "
                + str(MemInfo["SwapTotal"])
                + ","
                + str(MemInfo["SwapUsed"])
                + ", "
                + str(MemInfo["SwapFree"])
                + ","
                + str(MemInfo["Cached"])
                + ","
                + str(MemInfo["Buffers"])
                + ","
                + str(CpuInfo["Ncpu"])
                + ", "
                + str(CpuInfo["Usage"])
                + " , '"
                + gather
                + "' , "
                + str(timeNow)
                + ")"
            )

        ret = self.__eUtil.runUpdateSQL(sql)
        if ret is None or ret < 1:
            logger.info("+ServerMonitor.__updateResourceStatus() - host %s engine_monitoring update failed  sql = %s", self.hostname, sql)

    def __isRunningProcess(self, depID, myName="mainEngine"):
        """
        Method to return a list of process ids for current server monitor and child 'workmanager' processes.
        First list is the ones owned the current parent process and the second list is owned by a different process group.

        JDW -- remove exit condition for this method
        """

        myList = self.__pu.findProcesses(key="cmdline", value=myName, op="str_in")
        myPidList = [t[0] for t in myList]

        oList = self.__pu.findProcesses(key="cmdline", value=depID, op="str_in")

        retMe = []
        retOther = []

        if myList is None or len(myList) < 1:
            logger.info("ServerMonitor.__isRunningProcess : WARNING - Could not find processes ids for entry %r process %r", depID, myName)
            return retMe, retOther

        for oTup in oList:
            if oTup[1] in myPidList:
                retMe.append(oTup[0])
            else:
                retOther.append(oTup[0])
        return retMe, retOther

    def __wakeUpWorkflow(self):
        """
        For all sleeping WF we need review whether the wf_instance state has changed
        This could be slow and so I need to review this as we may have many 1000 of
        sleeping WF.  Probably OK for UI - how do we manage 'review' can we get this
          to poke something directly to the communication table.

        (thinking - In WFE - if communication = "sleeping" for a "closed(?) UI we could
                    could update communication to "wakeup"

        Note : I have put in a cludge for now : I wait 60 seconds for the WF to start running before
               checking to wake it up

        """

        timeNow = self.__timeStamp.getSecondsFromReference()

        sql = (
            "select communication.ordinal,wf_instance_last.dep_set_id,wf_class_file,communication.host from wf_instance_last,communication where wf_instance_last.dep_set_id = communication.dep_set_id and communication.host = '"  # noqa: E501
            + self.hostname
            + "' and communication.activity = 'SLEEPING' and wf_instance_last.inst_status not in ("
        )
        for valid in self.validSleep:
            sql = sql + "'" + str(valid) + "',"
        # Problem - when is a wakeup call stale ?
        #  We have an issue that if the server is off line we lose the wakeup.
        #  If we make this longer, then sleep/wake bounces
        #
        sql = sql[:-1] + ") and (" + str(timeNow) + " - wf_instance_last.status_timestamp) < 60"

        myList = self.__eUtil.runSelectSQL(sql)

        if myList is None or len(myList) == 0:
            return
        else:
            for row in myList:
                logger.info("+ServerMonitor.__wakeUpWorkflow : detected sleeping entry %r with new task %r\n", row[1], row[2])
                # wait 1 minute to get the process started
                if row[1] in self.monitorDict:
                    m = self.monitorDict[row[1]]
                    if m["status"] == "sleeping":
                        logger.info("+ServerMonitor.__wakeUpWorkflow : checking monitored sleeping entry %r with new task %r\n", row[1], row[2])
                        pid, _subPid = self.__isRunningProcess(row[1], myName=self.__baseEngineProcessName)
                        if pid is None or len(pid) == 0:
                            m["status"] = "working"
                            command = "update communication set activity = 'WORKING', actual_timestamp = " + str(timeNow) + " where ordinal in (" + str(row[0]) + ")"
                            _nrow = self.__eUtil.runUpdateSQL(command)  # noqa: F841
                            # need to recover the annotation flow - so run the annotation-WF at the correct module-WF at the correct task
                            # there is no process so run it
                            # look up a workflow nesting of task
                            # if there is a workflow nesting, then force a ,looup from annotate root, otherwise just recover
                            n = row[2].find(":")
                            if n > 0:
                                row[2] = row[2][:n]
                                wf = WorkflowManager(row[2], row[1], "recoverannotate", "recover")
                            else:
                                wf = WorkflowManager(row[2], row[1], "recover", "recover")

                            logger.info("+ServerMonitor.__wakeUpWorkflow : recovered %s at %s", str(row[1]), str(row[2]))

                            pid = wf.runWF(wait=False)
                            # set the ID to the new ownership
                            m["pid"] = pid
                        else:
                            m["status"] = "working"
                            logger.info("+ServerMonitor.__wakeUpWorkflow : Error : workflow for %s is running - impossible ", str(row[1]))
                    else:
                        logger.info("+ServerMonitor.__wakeUpWorkflow Process is already working %s", str(row[1]))
                else:
                    # JDW --
                    logger.info("+ServerMonitor.__wakeUpWorkflow : ignoring sleeping entry %r with new task %r on host %s\n", row[1], row[2], row[3])

    def __showMonitored(self):
        """Log the current list of monitored entries"""
        logger.info("+ServerMonitor.__showMonitored  total monitored entries %d", len(self.monitorDict))
        for ky in self.monitorDict.keys():
            m = self.monitorDict[ky]
            logger.info("+++++      %-10s %-10s", m["depID"], m["status"])

    def __monitorRestart(self, verbose=False):
        """
        Method to recover the run time state of the monitoring server from data in the communication table

        1) Get all entries ASSIGNED to this host/server in WORKING or SLEEPING state
        2) If a workflow is working - restart with recovery, add the entry

        """
        if verbose:
            logger.info("+ServerMonitor.__monitorRestart STARTING ----------------------------------------------------  ")
            logger.info("+ServerMonitor.__monitorRestart INITIAL MONITOR LIST  ---------------------------------------")
            self.__showMonitored()

        sql = "select ordinal,dep_set_id,activity,wf_class_file from communication where host = '" + self.hostname + "' and activity in ('SLEEPING', 'WORKING' )"
        allList = self.__eUtil.runSelectSQL(sql)

        if verbose:
            logger.info("+ServerMonitor.__monitorRestart found candidate entries ASSIGNED to host %s = %s", str(self.hostname), str(len(allList)))

        for row in allList:
            #  - skip any monitored entries -
            if row[1] in self.monitorDict:
                continue
            _pidMe, pidOther = self.__isRunningProcess(row[1], myName=self.__baseEngineProcessName)
            if pidOther is not None:
                self.__pu.killProcessList(pidOther)
            #
            if row[2] == "WORKING":
                #
                n = row[3].find(":")
                if n > 0:
                    row[3] = row[3][:n]
                    wf = WorkflowManager(row[3], row[1], "recoverannotate", "recover")
                else:
                    wf = WorkflowManager(row[3], row[1], "recover", "recover")

                logger.info("+ServerMonitor.__monitorRestart : recovering WORKING entry %s with activity %s", str(row[1]), str(row[2]))
                pid = wf.runWF(wait=False)
                m = {}
                m["depID"] = row[1]
                m["classID"] = row[3]
                m["pid"] = pid
                m["timeStamp"] = self.__timeStamp.getSecondsFromReference()
                m["status"] = "working"
                self.monitorDict[row[1]] = m
            elif row[2] == "SLEEPING":
                #
                m = {}
                logger.info("+ServerMonitor.__monitorRestart:  monitoring SLEEPING entry %s", str(row[1]))
                m["depID"] = row[1]
                m["classID"] = row[3]
                m["pid"] = None
                m["timeStamp"] = self.__timeStamp.getSecondsFromReference()
                m["status"] = "sleeping"
                self.monitorDict[row[1]] = m
        #
        if verbose:
            logger.info("+ServerMonitor.__monitorRestart REVISED MONITOR LIST  ---------------------------------------")
            self.__showMonitored()
            logger.info("+ServerMonitor.__monitorRestart COMPLETED  ---------------------------------------")
        #

    def __sleepWorkFlow(self):
        """
          Looks for idle ACTIVE (WORKING) entries with  instance status (waiting/open)


        method to review the current running annotations and look for WF that
        are not doing anything
        a) open or waiting UI   (wf_instance.status = 'waiting' or 'open')
        b) waiting for decision update from journal/depositor/annotator
                      (wf_instance = 'review') ????
        1) Open/waiting UI can be put to sleep
        2) Must be very careful what we put to sleep - processes and workflows cannot be slept - they are active.

        We just kill all processes associated with this WF if they have done
          nothing for at least 1 hour

        """

        timeNow = self.__timeStamp.getSecondsFromReference()

        # get all WF that are owned by this server that have 'working' status but have not changed state for "purgetime"

        sql = (
            "select communication.ordinal,wf_instance.dep_set_id from wf_instance,communication where wf_instance.dep_set_id = communication.dep_set_id and communication.host = '"
            + self.hostname
            + "' and communication.activity = 'WORKING' and wf_instance.inst_status in ("
        )
        for valid in self.validSleep:
            sql = sql + "'" + str(valid) + "',"

        sql = sql[:-1] + ") and (" + str(timeNow) + " - wf_instance.status_timestamp) > " + str(self.purgeTime)

        allList = self.__eUtil.runSelectSQL(sql)

        if allList is None or len(allList) < 1:
            return

        command = "update communication set activity = 'SLEEPING', actual_timestamp = " + str(timeNow) + " where ordinal in ("

        n = 0
        if allList is not None:
            for allRow in allList:
                if allRow[1] in self.monitorDict:
                    m = self.monitorDict[allRow[1]]
                    if m["status"] == "working":
                        logger.info("+ServerMonitor.__sleepWorkFlow : checking process list workflow for %s ", allRow[1])
                        pid, _other = self.__isRunningProcess(allRow[1], myName=self.__baseEngineProcessName)
                        if pid is not None:
                            m["status"] = "sleeping"
                            logger.info("+ServerMonitor.__sleepWorkFlow : killing processes for %s ", allRow[1])
                            self.__manageKillProcess(timeNow, 0, allRow)
                            command = command + str(allRow[0]) + ","
                            n = n + 1
                        else:
                            logger.info("+ServerMonitor.__sleepWorkFlow : no process associated with %r", allRow[1])
                            m["status"] = "sleeping"
                    else:
                        logger.info("+ServerMonitor.__sleepWorkFlow found unexpected workflow in sleeping  state for %r", allRow[1])
                        m["status"] = "sleeping"

        command = command[:-1] + ")"

        if n > 0:
            nrow = self.__eUtil.runUpdateSQL(command)
            logger.info("+ServerMonitor.__sleepWorkFlow  workflow sleeper - number of ID's put to sleep = %s", str(nrow))
        else:
            # JDW this update code is never applied so there are no DEAD workflows !
            command = "update communication set activity = 'DEAD', actual_timestamp = " + str(timeNow) + " where ordinal in ("
            p = ""
            for allRow in allList:
                #    logger.info("+ServerMonitor.sleepWorkFlow  ID = " + str(allRow[1]))
                command = command + str(allRow[0]) + ","
                p = p + str(allRow[0]) + ","
            command = command[:-1] + ")"
            # logger.info("+ServerMonitor.sleepWorkFlow  Warning : Stale working deposition : No process found to kill for now " + str(p))
            #      nrow = self.__eUtil.runUpdateSQL(command)

    def __getStatusData(self, inStatus, outStatus):  # pylint: disable=unused-argument
        """
        Returns a list (1) of dictionaries describing PENDING workflows TO BE monitored by this system.

        This method ASSIGNS hostname in the communication table.

        inStatus: is the target command in the communication table (e.g runWF)

        outStatus:  NOT USED

        Returns:  data ONLY from the communications table for the selected PENDING entry -

        """

        # Take the rowLimit=1 list of pending tasks in communication table with command = inStatus
        #
        command = (
            "select communication.ordinal, communication.dep_set_id, communication.wf_class_ID, communication.wf_inst_id,communication.wf_class_file, communication.sender, communication.actual_timestamp, parent_wf_class_id from communication where communication.receiver = 'WFE' and communication.command = '"  # noqa: E501
            + inStatus
            + "' and communication.status = 'PENDING' limit "
            + str(self.rowLim)
        )

        myList = self.__eUtil.runSelectSQL(command)

        if len(myList) < 1:
            return {}

        # just to make sure we capture multiple pending for one ID

        timeNow = self.__timeStamp.getSecondsFromReference()
        tS = self.__timeStamp.getTimeStringLocal(timeNow)
        me = self.hostname
        #
        # This will always return something -  command = "update communication set
        # activity = 'WORKING', status = 'WORKING', host = '" + me + "',
        # actual_timestamp = " + str(timeNow) + " where ordinal in ("

        # this will only update a row if the status was pending - and this server obtained the work - all server make the same generic request
        command = "update communication set activity = 'WORKING', status = 'WORKING'  where ordinal in ("
        for row in myList:
            command = command + str(row[0]) + ","
        command = command[:-1] + ")"

        nrow = self.__eUtil.runUpdateSQL(command)
        logger.info("+SeverMonitor.getStatusData updated %d rows with WORKING status ordinals %r at %s", nrow, myList, tS)

        # Logic only works for limit = 1 for now, if we updated no rows, then we did not get ownership of rows.
        if nrow == 0:
            return []
        else:
            ret = []
            # now we own job - mark the host and time stamp
            command = "update communication set host = '" + me + "', actual_timestamp = " + str(timeNow) + " where ordinal in ("
            for row in myList:
                command = command + str(row[0]) + ","
            command = command[:-1] + ")"
            nrow = self.__eUtil.runUpdateSQL(command)
            if nrow > 0:
                for row in myList:
                    rdict = {}
                    rdict["DEP_SET_ID"] = row[1]
                    rdict["WF_INST_ID"] = row[3]
                    rdict["WF_CLASS_FILE"] = row[4]
                    # class_id or parent_class_id
                    if row[2]:
                        rdict["WF_CLASS_ID"] = row[2]
                    else:
                        rdict["WF_CLASS_ID"] = row[7]
                    ret.append(rdict)
                    logger.info("+SeverMonitor.getStatusData returns dict %r ", rdict.items())
                    break
            else:
                logger.info("+ServerMonitor.getStatusData : Failed to get the job %s", str(command))
                ret = []
            return ret

    def __getWorkStep(self):

        load = self.__checkProcess()

        if load > self.normalLoad:
            if load > self.extremeLoad:
                return 3
            else:
                return 2
        else:
            return 1

    # --------------------------------------------------------------------------------------------------------------
    def XwatchList(self, serverList):
        """
        Get the list of wf-instances that should be active
        running - check the serverList to see if anyone owns this - it should be active
        closed(?) - this is waiting to be picked up by a server - does a server own this ?
        waiting/open are valid states for states that are not managed by a server
        """

        # get the list of ID which are running or closed
        sql = "select distinct(dep_set_id) from wf_instance where substr(inst_status,1,4) in ('runn','clos') and substr(dep_set_id,1,2) = 'D_'"

        myList = self.__eUtil.runSelectSQL(sql)

        if len(myList) == 0:
            # there are no ID running or closed
            return
        else:
            for row in myList:
                # check the server list to see if someone owns this entry
                found = False
                for serverID in serverList:
                    if serverID[2].find(row[0]) >= 0:
                        found = True
                        break
                if not found:
                    # No server owns this deposition
                    #          print " No server owns dep_set_id = " + str(row[0])
                    sql = (
                        "select wf_class_id,wf_inst_id,inst_status,status_timestamp,owner from wf_instance where dep_set_id = '"
                        + row[0]
                        + "' and wf_class_id <> 'popDB' order by status_timestamp desc limit 1"
                    )
                    watch = self.__eUtil.runSelectSQL(sql)
                    if watch is not None:
                        for row2 in watch:
                            mode = 0
                            if row2[2].find("clos") >= 0:
                                # we have found a closed dep_set_id
                                logger.info("Found a closed UI %s, %s, %s", str(row[0]), str(row2[0]), str(row2[1]))
                                mode = 1
                            elif row2[2].find("runn") >= 0:
                                logger.info("Found a running dep %s, %s, %s", str(row[0]), str(row2[0]), str(row2[1]))
                                mode = 2
                            # get the current task for this instance
                            if mode > 0:
                                sql = "select wf_task_id from wf_task where dep_set_id = '" + row[0] + "' and wf_inst_id = '" + row2[1] + "' order by status_timestamp desc limit 1"
                                tasks = self.__eUtil.runSelectSQL(sql)
                                taskID = None
                                if tasks is not None:
                                    for task in tasks:
                                        taskID = task[0]
                                        logger.info("        current task = %s", taskID)
                                if taskID is not None:
                                    m = {}
                                    # At this stage we can see that no server has access to this process.
                                    # we need to see if the process is actually running already
                                    logger.info("running workflow to capture orphen job")
                                    pids, _subPid = self.__isRunningProcess(row[0], myName=self.__baseEngineProcessName)
                                    if pids is None or len(pids) < 1:
                                        # there is no process so run it
                                        wf = WorkflowManager(row2[4], row[0], None, taskID)
                                        pid = wf.runWF(wait=False)
                                        m["pid"] = pid
                                    else:
                                        if len(pids) == 1:
                                            m["pid"] = pids[0]
                                        else:
                                            #  we need to find the parent
                                            m["pid"] = pids[0]
                                            logger.info(" Oh crap - I got multiple ID ")
                                    m = {}
                                    m["depID"] = row[0]
                                    m["classID"] = row2[4]
                                    m["timeStamp"] = self.__timeStamp.getSecondsFromReference()
                                    m["status"] = "working"
                                    self.monitorDict[row[0]] = m



##
# File:  WFEngineRunner.py
# Date:  13-Feb-2015  J.Westbrook
# Updates:
#
###
##
"""
Launch and restart workflow engine process.

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) 2010-2015 wwPDB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import os
import sys
import logging

from wwpdb.apps.wf_engine.engine.mainEngine import mainEngine
from wwpdb.utils.detach.DetachedProcessBase import DetachedProcessBase


logger = logging.getLogger(name="root")


class WFEngineRunner(DetachedProcessBase):
    """Launch and restart workflow engine process."""

    def __init__(self, pidFile="WFEngineRunner.pid", debugLevel=1, wfXmlPath=".", stdin=os.devnull, stdout=os.devnull, stderr=os.devnull, wrkDir="."):

        super(WFEngineRunner, self).__init__(pidFile=pidFile, stdin=stdin, stdout=stdout, stderr=stderr, wrkDir=wrkDir)

        self.__debugLevel = debugLevel
        self.__pidFile = pidFile
        self.__wfXmlPath = wfXmlPath

    def run(self):
        """Start a new workflow engine process  -"""
        logger.info("+WFEngingeRunner.run() Start workflow engine  with %s\n", self.__pidFile)

        engine = mainEngine(self.__debugLevel, sys.stderr)
        normal = ["WFRunner", "-x", "-k", "WF", "-t", "entry-point", "-s", "monitor", "-w", "MonitorDB.xml", "-p", self.__wfXmlPath]
        engine.runNoThrow(normal)

        logger.info("+WFEngineRunner.run() Leaving with %s\n", self.__pidFile)
        return True

    def setDebugLevel(self, level=0):
        self.__debugLevel = level


if __name__ == "__main__":
    wfr = WFEngineRunner()
    wfr.run()



##
# File: TaskParameter
# Date: 2-May-2015
#
# #
class TaskParameter(object):
    """Container of task parameter details"""

    def __init__(self):

        self.title = ""
        self.format = ""
        self.comment = ""
        self.question = ""
        self.options = []
        self.objects = []

    def addFormat(self, string):

        self.format = string

    def addComment(self, string):

        self.comment = string

    def addQuestion(self, string):

        self.question = string

    def addTitle(self, string):

        self.title = string

    def addOption(self, n, string):

        self.options.append([str(n), string])

    def addObject(self, string):

        self.objects.append(string)



##
# File:    WorkflowSession.py
# Date:    15-Mar-2009
#
# Updates:
#  10-May-2010 : convert to use the status DB
#
#  2-May-2015  jdw -- Refactor -- sql and db connections
##

"""
  class to manage recovery from old workflow states
"""

import sys
import logging
from wwpdb.apps.wf_engine.wf_engine_utils.time.TimeStamp import TimeStamp

logger = logging.getLogger(name="root")


class WorkflowSession(object):
    def __init__(self, eUtil, depID, classID, debug=True, prt=sys.stderr):  # pylint: disable=unused-argument
        self.depID = depID
        self.classID = classID
        self.__eUtil = eUtil
        self.__timeStamp = TimeStamp()
        self.debug = debug
        self.taskID = None
        # self.__lfh = prt

    def autoRecover(self):

        instID = self.__getLastInstanceID()

        if instID is None:
            return None, None

        taskID = self.autoRecoverInstance(instID)

        return instID, taskID

    def autoRecoverInstance(self, instID):
        """
        Reads the status DB to to find the last task for an instance
        to see where it died
        Finds the last task that did NOT finish
        """

        sql = (
            "select wf_task_id,task_name,task_type,task_status,status_timestamp from wf_task where dep_set_id = '"
            + self.depID
            + "' and wf_class_id = '"
            + self.classID
            + "' and wf_inst_id = '"
            + instID
            + "' and not task_status = 'finished' order by status_timestamp limit 1 "
        )

        ret = self.__eUtil.runSelectSQL(sql)

        # we now should have 1 row of task =
        # the last one that finished

        if ret is not None:
            for row in ret:
                logger.info("+WorkflowSession.autoRecover :  Last not finished = %s, of task type = %s completed at %s", str(row[1]), str(row[2]), str(row[4]))
                self.taskID = str(row[0])
                return row[0]
        else:
            logger.info("+WorkflowSession.getLastInstanceID : WF error - cannot recover this WF - no task ever finished ")
            self.taskID = "entry-point"
            return None

    def recoverFromAnnotate(self):

        sql = "select wf_class_id from wf_instance_last where dep_set_id = '" + str(self.depID) + "'"
        ss = self.__eUtil.runSelectSQL(sql)
        for s in ss:
            taskID = s[0]

        return taskID

    def autoRecoverInstanceGetLastFinished(self, instID):
        """
        Reads the status DB to to find the last task for an instance
        to see where it died
        Finds the last task that finished, and recovers the
          reference data for that task, and restarts at the
          next task
        """

        sql = (
            "select wf_task_id,task_name,task_type,task_status,status_timestamp from wf_task where dep_set_id = '"
            + self.depID
            + "' and wf_class_id = '"
            + self.classID
            + "' and wf_inst_id = '"
            + instID
            + "' and task_status = 'finished' order by status_timestamp desc limit 1"
        )

        ret = self.__eUtil.runSelectSQL(sql)

        # we now should have 1 row of task =
        # the last one that finished

        if ret is not None:
            for row in ret:
                logger.info("+WorkflowSession.autoRecover :  Last finished = %s, of task type = %s completed at %s", str(row[1]), str(row[2]), str(row[4]))
                self.taskID = str(row[0])
                return row[0]
        else:
            logger.info("+WorkflowSession.getLastInstanceID : WF error - cannot recover this WF - no task ever finished ")
            self.taskID = "entry-point"
            return None

    def __getLastInstanceID(self):
        """
        Find that last occurance of the instanceID
        """

        sql = (
            "select wf_inst_id,status_timestamp from wf_instance where dep_set_id = '" + self.depID + "' and wf_class_id = '" + self.classID + "' order by wf_inst_id desc limit 1"
        )

        if self.debug:
            logger.info("+WorkflowSession.__getLastInstanceID :  looking for last instance for depID= %s, classID= %s", str(self.depID), str(self.classID))

        ret = self.__eUtil.runSelectSQL(sql)

        if ret is not None and len(ret) > 0:
            for row in ret:
                instID = row[0]
                delta = float(row[1] - self.__timeStamp.getSecondsFromReference()) / 60.0
                logger.info("+WorkflowSession.__getLastInstanceID : Found instID = %s  age = %.3f minutes", instID, delta)
            return instID
        else:
            logger.info("+WorkflowSession.__getLastInstanceID : Catastrophic WF error - cannot recover instance ID ")
            return None

    # def __recoverAt(self, instID, taskID):
    #     """  NOT USED
    #       Reads the status DB to to find the last inst_ID
    #       to see where it died
    #       Recover from the prescribed taskID
    #     """

    #     sql = "select wf_task_id,task_name,task_type,task_status,status_timestamp from wf_task where dep_set_id = '" + self.depID + "' and wf_class_id = '" + \
    #         self.classID + "' and wf_inst_id = '" + instID + "' and wf_inst_id = '" + taskID + "'  order by status_timestamp desc limit 1"

    #     ret = self.__eUtil.runSelectSQL(sql)

    #     # we now should have 1 row of task =
    #     # the last one that finished

    #     if ret is not None:
    #         for row in ret:
    #             logger.info("+WorkflowSession.autoRecover :  Last finished = %s, of task type %s completed at %s", str(row[1]), str(row[2]), str(row[4]))
    #             return row[0]
    #     else:
    #         logger.info("+WorkflowSession.getLastInstanceID : WF error - cannot recover this WF - no task ever finished ")
    #         return "entry-point"



##
# File:    WorkflowManager.py
# Date:    15-Mar-2015
#
# Updates:
#   20-Mar-2015   jdw cleanup all of the hardcoded and inconsistent pathing
#                     move logfiles out data directories --
#
#
##
import time
import os
import threading
import subprocess
import logging
from wwpdb.utils.config.ConfigInfo import ConfigInfo, getSiteId

logger = logging.getLogger(name="root")


class WorkflowManager(threading.Thread):
    def __init__(self, file, depID, instID, taskID="entry-point"):
        threading.Thread.__init__(self)
        self.status = 0
        self.file = file

        self.depID = depID
        self.instID = instID
        self.taskID = taskID
        self.__siteId = getSiteId(defaultSiteId="WWPDB_DEPLOY_TEST")
        self.__cI = ConfigInfo(self.__siteId)
        self.__wfXmlPath = self.__cI.get("SITE_WF_XML_PATH")

    # def setCommand(self, command):
    #     self.command = command

    def abort(self):
        self.status = 8

    def runWF(self, wait):

        self.status = 1

        logDir = self.__getLogDirectoryPath(self.depID)
        logFile = os.path.join(logDir, str(self.depID) + "_WF_" + str(self.file[:-4]) + ".log")
        logger.info("+WorkflowManager.runWF() -------------------------------------------------------")
        logger.info("+workflowManager.runWF() :    siteID       = %s", str(self.__siteId))
        logger.info("+workflowManager.runWF() :     depID       = %s", str(self.depID))
        logger.info("+workflowManager.runWF() :  WF class file  = %s", str(self.file))
        logger.info("+workflowManager.runWF() :   taskID        = %s", str(self.taskID))
        logger.info("+workflowManager.runWF() :  XML path       = %s", self.__wfXmlPath)
        logger.info("+workflowManager.runWF() :   logfile       = %s", str(logFile))
        logger.info("+workflowManager.runWF() : wait flag       = %s", str(wait))

        if self.instID is None:
            args = [
                "python",
                "-m",
                "wwpdb.apps.wf_engine.engine.mainEngine",
                "-0",
                "-s",
                self.depID,
                "-t",
                self.taskID,
                "-d",
                "2",
                "-l",
                logFile,
                "-w",
                self.file,
                "-p",
                self.__wfXmlPath,
            ]
        else:
            args = [
                "python",
                "-m",
                "wwpdb.apps.wf_engine.engine.mainEngine",
                "-0",
                "-s",
                self.depID,
                "-i",
                self.instID,
                "-t",
                self.taskID,
                "-d",
                "2",
                "-l",
                logFile,
                "-w",
                self.file,
                "-p",
                self.__wfXmlPath,
            ]

        istat = subprocess.Popen(args)
        if wait:
            istat.wait()

        self.status = 2
        time.sleep(1)
        logger.info("+WorkflowManager.runWF() workflow subprocess for %s task %r with pid %r", self.depID, self.taskID, str(istat))
        logger.info("+WorkflowManager.runWF() -------------------------------------------------------")
        return istat

    def run(self):

        self.status = 1
        logDir = self.__getLogDirectoryPath(self.depID)
        logFile = os.path.join(logDir, str(self.depID) + "_WF_" + str(self.file[:-4]) + ".log")
        logger.info("+workflowManager.run() :   siteID      = %s", str(self.__siteId))
        logger.info("+WorkFlowManager.run() :   depID       = %s", str(self.depID))
        logger.info("+WorkFlowManager.run() : WF class file = %s", str(self.file))
        logger.info("+WorkFlowManager.run() :  taskID       = %s", str(self.taskID))
        logger.info("+WorkFlowManager.run() : XML path      = %s", self.__wfXmlPath)
        logger.info("+WorkFLowManager.run() : logfile       = %s", str(logFile))

        if self.instID is None:
            args = [
                "python",
                "-m",
                "wwpdb.apps.wf_engine.engine.mainEngine",
                "-0",
                "-s",
                self.depID,
                "-t",
                self.taskID,
                "-d",
                "2",
                "-l",
                logFile,
                "-w",
                self.file,
                "-p",
                self.__wfXmlPath,
            ]

        else:
            args = [
                "python",
                "-m",
                "wwpdb.apps.wf_engine.engine.mainEngine",
                "-0",
                "-s",
                self.depID,
                "-i",
                self.instID,
                "-t",
                self.taskID,
                "-d",
                "2",
                "-l",
                logFile,
                "-w",
                self.file,
                "-p",
                self.__wfXmlPath,
            ]

        istat = subprocess.call(args)
        if istat == 0:
            self.status = 2
        else:
            self.status = 3
        time.sleep(1)
        #
        logger.info("+WorkflowManager.run() workflow subprocess for %s task %r returns value %r ", self.depID, str(self.taskID), str(istat))
        logger.info("+WorkflowManager.run() -------------------------------------------------------")
        return

    def workflowStatus(self):
        return self.status

    def __getLogDirectoryPath(self, depID):

        topSessionPath = self.__cI.get("SITE_WEB_APPS_TOP_SESSIONS_PATH")
        logDir = os.path.join(topSessionPath, "wf-logs", depID)

        if not os.path.exists(logDir):
            os.makedirs(logDir)

        return logDir



##
# File: ProcessUtils.py
# Date: 14-Feb-2015  J. Westbrook --
#
# Update:
# 12-Mar-2015 jdw add process search method ()
# 12-Mar-2015 jdw return parent process id in search method
#
##

__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.01"

import os
import psutil
import signal
import traceback


class ProcessUtils(object):
    """
    A collection of utilities for managing processes and extracting
    process and system status details.

    """

    def __init__(self, verbose, log):  # pylint: disable=unused-argument
        # self.__verbose = verbose
        self.__lfh = log
        self.__debug = False

    def setDebug(self, flag=True):
        self.__debug = flag

    def __compare(self, value, target, op="str_in"):
        """
        Compare value with target according to input operation.

        This is tailored to data types returned from process status queries.

        Return True if the comparison is satisfied
        """
        ok = False
        opList = ["str_in", "str_eq", "numb_eq", "numb_gt", "numb_lt"]
        if op not in opList:
            return False
        if target is None:
            return False
        if op in ["str_in"]:
            if isinstance(target, list):
                ok = value in " ".join(target)
            elif isinstance(target, str):
                ok = value in target
        elif op in ["str_eq"]:
            if isinstance(target, list):
                ok = value in target
            elif isinstance(target, str):
                ok = value == target
        elif op in ["numb_gt"]:
            ok = value < target
        elif op in ["numb_lt"]:
            ok = value > target
        elif op in ["numb_eq"]:
            ok = value == target
        else:
            ok = False
        return ok

    def findProcesses(self, key, value, op="str_in"):
        """Return the list of process & parent ids in which input value is found with the process target key -
        Queries are performed on the current process list.  Comparisons/selections performed with
        self.__compare() according the operation selection.
        """
        # valid list of process status  - keys -
        keyList = [
            "username",
            "num_ctx_switches",
            "pid",
            "connections",
            "cmdline",
            "create_time",
            "memory_info_ex",
            "num_fds",
            "memory_maps",
            "cpu_percent",
            "terminal",
            "ppid",
            "cwd",
            "nice",
            "status",
            "cpu_times",
            "memory_info",
            "threads",
            "open_files",
            "name",
            "num_threads",
            "exe",
            "uids",
            "gids",
            "memory_percent",
        ]
        pidList = []
        if key not in keyList:
            return pidList
        for p in psutil.process_iter():
            try:
                # jdw  - It is observed that processes may processes may not be available to
                #        to this call so trap this exception so that the iterator can continue.
                try:
                    pD = p.as_dict()
                except:  # noqa: E722 pylint: disable=bare-except
                    pass
                if key in pD:
                    # if self.__debug:
                    #    self.__lfh.write("+ProcessUtils.findProcess() testing for key %r  search target %r process return value %r\n" % (key, value, pD[key]))
                    if pD[key] is None:
                        continue
                    ok = self.__compare(value, pD[key], op=op)
                    if ok:
                        pidList.append((pD["pid"], pD["ppid"]))
                        if self.__debug:
                            self.__lfh.write("+ProcessUtils.findProcess() matched key %r value %r\n" % (key, pD[key]))
            except:  # noqa: E722 pylint: disable=bare-except
                if self.__debug:
                    self.__lfh.write("+ProcessUtils.findProcess() exception for key %r target %r\n" % (key, value))
                    traceback.print_exc(file=self.__lfh)
        if self.__debug:
            self.__lfh.write("+ProcessUtils.findProcess() key %r target %r search returns length %d\n" % (key, value, len(pidList)))
        return pidList

    def killProcessList(self, pidList, mySignal=signal.SIGKILL):
        try:
            for pid in pidList:
                os.kill(pid, mySignal)

                try:
                    os.waitpid(pid, 0)
                    self.__lfh.write("+ProcessUtils.killProcessList() process %s cleaned from process table\n" % (str(pid)))
                except OSError as e:
                    self.__lfh.write("+ProcessUtils.killProcessList() error waiting for pid (%s), error %s\n" % (str(pid), e))
            return True
        except Exception as e:
            self.__lfh.write("+ProcessUtils.killProcessList() failing with (%s)\n" % (str(e)))
        return False

    def getChildren(self, pid):
        """Return the list of process id's for the children of the input process id."""
        cL = []
        try:
            p = psutil.Process(pid)
            cPidList = p.children(recursive=True)
            for cPid in cPidList:
                iPid = int(str(cPid.pid))
                if iPid != pid:
                    cL.append(iPid)
        except Exception as e:
            self.__lfh.write("+ProcessUtils.getPidChildren() failing with (%s)\n" % (str(e)))

        return cL

    def getMemoryInfo(self):
        """
        From the  psutil class documentation -

        virtual_memory() returns statistics about system memory usage as a namedtuple including the following fields, expressed in bytes:

        total: total physical memory available.
        available: the actual amount of available memory that can be given instantly to processes that request more memory in bytes; this
        is calculated by summing different memory values depending on the platform (e.g. free + buffers + cached on Linux) and it is supposed
         to be used to monitor actual memory usage in a cross platform fashion.
        percent: the percentage usage calculated as (total - available) / total * 100.
        used: memory used, calculated differently depending on the platform and designed for informational purposes only.
        free: memory not being used at all (zeroed) that is readily available; note that this does not reflect the actual memory
                   available (use available instead).
        Platform-specific fields:

        active: (UNIX): memory currently in use or very recently used, and so it is in RAM.
        inactive: (UNIX): memory that is marked as not used.
        buffers: (Linux, BSD): cache for things like file system metadata.
        cached: (Linux, BSD): cache for various things.
        wired: (BSD, OSX): memory that is marked to always stay in RAM. It is never moved to disk.
        shared: (BSD): memory that may be simultaneously accessed by multiple processes.

        swap_memory() returns system swap memory statistics as a namedtuple including  the following fields:

            total: total swap memory in bytes
            used: used swap memory in bytes
            free: free swap memory in bytes
            percent: the percentage usage
            sin: the number of bytes the system has swapped in from disk (cumulative)
            sout: the number of bytes the system has swapped out from disk (cumulative)
        """
        d = {}
        d = {"MemTotal": 0, "SwapTotal": 0, "TotalFree": 0, "SwapFree": 0, "TotalUsed": 0, "Cached": 0, "MemFree": 0, "Buffers": 0, "TotalTotal": 0, "MemUsed": 0, "SwapUsed": 0}
        try:
            vM = psutil.virtual_memory()._asdict()
            sM = psutil.swap_memory()._asdict()
            #
            d["MemTotal"] = int(vM["total"] / 1000)
            d["MemUsed"] = int(vM["used"] / 1000)
            d["MemFree"] = int(vM["available"] / 1000)
            if "buffers" in vM:
                d["Buffers"] = int(vM["buffers"])
            #
            if "cached" in vM:
                d["Cached"] = int(vM["cached"])
            #
            d["SwapTotal"] = int(sM["total"] / 1000)
            d["SwapUsed"] = int(sM["used"] / 1000)
            d["SwapFree"] = int(sM["free"] / 1000)
            #
            d["TotalUsed"] = d["MemUsed"] + d["SwapUsed"]
            d["TotalFree"] = d["MemFree"] + d["SwapFree"]
            d["TotalTotal"] = d["MemTotal"] + d["SwapTotal"]

        except Exception as e:
            self.__lfh.write("+ProcessUtils.getMemoryInfo() failing with (%s)\n" % (str(e)))
            traceback.print_exc(file=self.__lfh)
        return d

    def getCpuInfo(self, waitSeconds=1.0):
        """Return the number of cores and aggregate fractional usage during the input
        interval (waitSeconds).

        """
        d = {}
        d["Ncpu"] = 1
        d["Usage"] = 0.001
        try:
            # total number of cores --
            d["Ncpu"] = psutil.cpu_count(logical=True)
            # aggregate fractional usage -
            d["Usage"] = float(psutil.cpu_percent(interval=waitSeconds)) / 100.0
        except Exception as e:
            self.__lfh.write("+ProcessUtils.getCpuInfo failing with (%s)\n" % (str(e)))
            traceback.print_exc(file=self.__lfh)

        return d



##
# File:    parseXML.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#
##

"""
entry point for the workflow class XML parser
"""

__docformat__ = "restructuredtext en"
__author__ = "Tom Oldfield"
__email__ = "oldfield@ebi.ac.uk"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.24"

import sys
from wwpdb.apps.wf_engine.reader.taskReference import taskModule
from wwpdb.apps.wf_engine.reader.dataReference import dataModule
from wwpdb.apps.wf_engine.reader.metaDataObject import metaDataObject


class parseXML(object):
    """Hold a data reference, passed variable is the XML dom
    object to be parsed"""

    def __init__(self, debug=0, prt=sys.stderr):

        self.debug = debug
        self.__lfh = prt

    def getMetaData(self, xmldoc):

        mData = None

        for outer in xmldoc.childNodes:
            if outer.nodeName == "wf:wwPDBworkflow":
                for thing in outer.childNodes:
                    if thing.nodeName == "wf:metadata":
                        mData = metaDataObject()
                        for metaData in thing.childNodes:
                            if metaData.nodeName == "wf:version":
                                self.__lfh.write(
                                    "WFE.parseXML.getMetaData : XML style = PRODUCTION : major release = "
                                    + str(metaData.getAttribute("major"))
                                    + " : Date = "
                                    + str(metaData.getAttribute("date"))
                                    + "\n"
                                )
                                mData.setVersionMajor(metaData.getAttribute("major"))
                                mData.setVersionMinor(metaData.getAttribute("minor"))
                                mData.setAuthor(metaData.getAttribute("author"))
                                mData.setDate(metaData.getAttribute("date"))
                                mData.setID(metaData.getAttribute("id"))
                                mData.setName(metaData.getAttribute("name"))
                            if metaData.nodeName == "wf:description":
                                for title in metaData.childNodes:
                                    if title.nodeName == "wf:short":
                                        mData.setDescription(title.firstChild.data)
                                    if title.nodeName == "wf:subtext":
                                        self.__lfh.write("WFE.parseXML.getMetaData :   --> " + str(title.firstChild.data) + "\n")

        return mData

    def getTaskObjects(self, xmldoc):

        taskObjects = []

        for outer in xmldoc.childNodes:
            for typeNode in outer.childNodes:
                if typeNode.nodeName == "wf:workflow":
                    for flowNodes in typeNode.childNodes:
                        if flowNodes.nodeName == "wf:flow":
                            for tasksNode in flowNodes.childNodes:
                                if tasksNode.nodeName == "wf:entryPoint":
                                    taskRef = taskModule(self.debug, self.__lfh)
                                    taskRef.type = "Entry-point"
                                    taskRef.name = tasksNode.getAttribute("taskID")
                                    taskRef.nameHumanReadable = tasksNode.getAttribute("name")
                                    taskRef.outputName.append(tasksNode.getAttribute("nextTask"))
                                    taskRef.exception = tasksNode.getAttribute("exceptionID")
                                    taskObjects.append(taskRef)
                                elif tasksNode.nodeName == "wf:exitPoint":
                                    taskRef = taskModule(self.debug, self.__lfh)
                                    taskRef.type = "Exit-point"
                                    taskRef.name = tasksNode.getAttribute("taskID")
                                    taskRef.nameHumanReadable = tasksNode.getAttribute("name")
                                    taskObjects.append(taskRef)
                                elif tasksNode.nodeName == "wf:exception":
                                    taskRef = taskModule(self.debug, self.__lfh)
                                    taskRef.type = "Exception"
                                    taskRef.parseXML(tasksNode)
                                    taskObjects.append(taskRef)
                                elif tasksNode.nodeName == "wf:tasks":
                                    for taskNode in tasksNode.childNodes:
                                        if taskNode.nodeName == "wf:task":
                                            taskRef = taskModule(self.debug, self.__lfh)
                                            taskRef.parseXML(taskNode)
                                            taskObjects.append(taskRef)
                                elif tasksNode.nodeName == "#text" or tasksNode.nodeName == "#comment":
                                    pass
                                else:
                                    self.__lfh.write("WFE.parseXML.getMetaData :  Critical workflow XMl error : Unknown FLOW tag " + str(tasksNode.nodeName) + "\n")
                                    sys.exit(0)

        return taskObjects

    def getDataObjects(self, depositionID, instanceID, xmldoc):

        dataObjects = []

        for outer in xmldoc.childNodes:
            for typeNode in outer.childNodes:
                if typeNode.nodeName == "wf:workflow":
                    for dataNodes in typeNode.childNodes:
                        if dataNodes.nodeName == "wf:dataObjects":
                            for dataNode in dataNodes.childNodes:
                                if dataNode.nodeName == "wf:dataObject":
                                    dat = dataModule(instanceID, depositionID, self.debug, self.__lfh)
                                    dat.parseXML(dataNode)
                                    dataObjects.append(dat)

        return dataObjects



##
# File:  WFTaskRequest.py
# Date:  10-April-2014  J. Westbrook
#
# Updates:
#
# 27-Jul-2015  add loadWfDefinition()
###
##
""" Manage workflow task requests and status queries --

This software was developed as part of the World Wide Protein Data Bank
Common Deposition and Annotation System Project

Copyright (c) 2010-2015 wwPDB

This software is provided under a Creative Commons Attribution 3.0 Unported
License described at http://creativecommons.org/licenses/by/3.0/.

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.07"

import os
import sys
import traceback
import platform
from wwpdb.utils.config.ConfigInfo import ConfigInfo
from wwpdb.utils.db.WorkflowSchemaDef import WorkflowSchemaDef
from wwpdb.utils.db.MyDbAdapter import MyDbAdapter
from wwpdb.apps.wf_engine.wf_engine_utils.time.TimeStamp import TimeStamp

#
from xml.dom import minidom
from wwpdb.apps.wf_engine.reader.parseXML import parseXML


class WFTaskRequest(MyDbAdapter):

    """Manage workflow task requests and status queries --"""

    def __init__(self, siteId, verbose=False, log=sys.stderr):
        super(WFTaskRequest, self).__init__(schemaDefObj=WorkflowSchemaDef(verbose=verbose, log=log), verbose=verbose, log=log)
        # self.__verbose = verbose
        self.__lfh = log
        self.__siteId = siteId
        self.__cI = ConfigInfo(self.__siteId)
        #
        self.__myFullHostName = platform.uname()[1]
        #
        self.__timeStamp = TimeStamp()
        self.__setup()

    def setDebug(self, flag=True):
        self._setDebug(flag=flag)

    def __setup(self):
        self.__commTableId = "COMMUNICATION"
        self.__commContextId = "COMMUNICATION"

        defValD = {
            "depSetId": "D_0000000000",
            "sender": "WFUTILS",
            "receiver": "WFE",
            "command": "runWF",
            "status": "PENDING",
            "wfInstId": None,
            "wfClassId": None,
            "wfClassFile": None,
            "actualTimestamp": None,
            "host": None,
            "activity": None,
            "dataVersion": "latest",
        }
        defValD["hostName"] = self.__myFullHostName
        self._setParameterDefaultValues(contextId=self.__commContextId, valueD=defValD)
        self.__wfTableIdList = ["COMMUNICATION", "WF_TASK", "WF_INSTANCE", "WF_INSTANCE_LAST", "DEPOSITION", "USER_DATA", "ENGINE_MONITORING", "WF_CLASS_DICT"]
        for tableId in self.__wfTableIdList:
            mapL = self._getDefaultAttributeParameterMap(tableId)
            self._setAttributeParameterMap(tableId=tableId, mapL=mapL)

        #
        # mapL = [('DEP_SET_ID', 'depId'),
        #        ('PARENT_WF_INST_ID', 'wfInstId'),
        #        ('PARENT_WF_CLASS_ID', 'wfClassId'),
        #        ('PARENT_DEP_SET_ID', 'parentDepId'),
        #        ('ACTUAL_TIMESTAMP', 'timeStamp'),
        #        ('SENDER', 'sender'),
        #        ('RECEIVER', 'receiver'),
        #        ('STATUS', 'status'),
        #        ('COMMAND', 'command'),
        #        ('WF_CLASS_ID', 'wfClassId'),
        #        ('WF_INST_ID', 'wfInstId'),
        #        ('WF_CLASS_FILE', 'wfClassFile'),
        #        ('DATA_VERSION', 'dataVersion'),
        #        ('HOST', 'hostName'),
        #        ('ACTIVITY', 'activity')
        #        ]

        cMapL = [("DEP_SET_ID", "depSetId")]
        for tableId in self.__wfTableIdList[:-1]:
            self._setConstraintParameterMap(tableId=tableId, mapL=cMapL)
        #
        #
        self.__reportFiltersD = {
            "ACTUAL_TIMESTAMP": (self.__timeStamp, "getTimeStringLocal"),
            "STATUS_TIMESTAMP": (self.__timeStamp, "getTimeStringLocal"),
            "ORDINAL_ID": (self, "_formatIntVal"),
        }
        #

    def loadWfDefinition(self, wfFileName):
        """
        Load the workflow description in the input workflow definition in wf_class_dict table
        overwriting any existing definition.

        """
        wfFilePath = os.path.join(self.__cI.get("SITE_WF_XML_PATH"), wfFileName)
        if os.access(wfFilePath, os.R_OK):
            debugLevel = 0
            xmldoc = minidom.parse(wfFilePath)
            parse = parseXML(debugLevel, self.__lfh)
            wfMetaData = parse.getMetaData(xmldoc)
            wfOpts = {}
            wfOpts["wfClassId"] = wfMetaData.getID()
            wfOpts["classFile"] = wfFileName
            wfOpts["wfClassName"] = wfMetaData.getName()
            wfOpts["title"] = wfMetaData.getDescription()
            wfOpts["author"] = wfMetaData.getAuthor()
            wfOpts["version"] = wfMetaData.getVersionMajor() + wfMetaData.getVersionMinor()
            #
            _ok = self._deleteRequest(tableId="WF_CLASS_DICT", wfClassId=wfMetaData.getID())  # noqa: F841
            return self._insertRequest(tableId="WF_CLASS_DICT", contextId=None, **wfOpts)
        else:
            return False

    def setDataStore(self, dataStoreName):
        self._setDataStore(dataStoreName=dataStoreName)

    def createDataStore(self):
        """Create/recreate tables defined in the class workflow definition."""
        self._createSchema()

    def __filterReportData(self, dList, fD):
        """apply filters on the input list of dictionaries -"""
        for d in dList:
            for k, v in fD.items():
                if k in d:
                    d[k] = getattr(v[0], v[1])(d[k])

    def _formatIntVal(self, iVal):
        return "%10d" % iVal

    def selectDeposition(self, depSetId, tableId):
        return self._select(tableId, depSetId=depSetId)

    def updateDeposition(self, depSetId, **kw):
        if "depSetId" not in kw:
            kw["depSetId"] = depSetId
        return self._updateRequest(tableId="DEPOSITION", contextId=None, **kw)

    def insertDeposition(self, depSetId, **kw):
        if "depSetId" not in kw:
            kw["depSetId"] = depSetId
        return self._insertRequest(tableId="DEPOSITION", contextId=None, **kw)

    def getReport(self, reportType=None, depSetId=None):
        """Internal method to fetch report data of the report input type  and
        optional data set identifier.
        """
        tableId = reportType
        if tableId in self.__wfTableIdList:
            if depSetId is None:
                dList = self._select(tableId=tableId)
                self.__filterReportData(dList, self.__reportFiltersD)
                return dList
            else:
                dList = self._select(tableId=tableId, depSetId=depSetId)
                self.__filterReportData(dList, self.__reportFiltersD)
                return dList
        else:
            return []

    def deleteDataSet(self, depSetId, tableId):
        """Delete the input deposition data set from the input table."""
        return self._deleteRequest(tableId=tableId, depSetId=depSetId)

    def clearTable(self, tableId):
        """Clear the input table."""
        return self._deleteRequest(tableId=tableId)

    def addDataSet(self, depSetId, **kwargs):
        """
        Add and entry for the dataset for the input deposition data set.

        optional qualifiers:

                      sender="WFUTILS"
                      receiver="WFE"
                      command="runWF"
                      status="PENDING"
                      wfInstId=None
                      wfClassId=None
                      wfClassFile=None
                      actualTimestamp=None
                      host=None
                      activity=None
                      dataVersion=None

        """
        options = self._getParameterDefaultValues(contextId=self.__commContextId)
        options.update(kwargs)
        options["depSetId"] = depSetId
        options["actualTimestamp"] = self.__timeStamp.getSecondsFromReference()
        return self._insertRequest(tableId=self.__commTableId, contextId=self.__commContextId, **options)

    def assignTask(self, depSetId, **kwargs):
        """
        Request the next task for the input deposition data set.

        depId               << required

        optional qualifiers:

                      sender="WFUTILS"
                      receiver="WFE"
                      command="runWF"
                      status="PENDING"
                      wfInstId=None
                      wfClassId=None
                      wfClassFile=None
                      actualTimeStamp=None
                      host=None
                      activity=None
                      dataVersion=None

        """
        options = self._getParameterDefaultValues(contextId=self.__commContextId)
        options.update(kwargs)
        options["depSetId"] = depSetId
        options["actualTimestamp"] = self.__timeStamp.getSecondsFromReference()
        return self._updateRequest(tableId=self.__commTableId, contextId=self.__commContextId, **options)

    def loadAccessions(self, tableId, accessionType=None, filePath=None):
        try:
            mapL = self._getDefaultAttributeParameterMap(tableId)
            self._setAttributeParameterMap(tableId=tableId, mapL=mapL)
            #
            idName = ""
            if accessionType == "PDB":
                cMapL = [("PDB_ID", "pdbId")]
                idName = "pdbId"
            elif accessionType == "BMRB":
                cMapL = [("BMRB_ID", "bmrbId")]
                idName = "bmrbId"
            elif accessionType == "EMDB":
                cMapL = [("EMDB_ID", "emdbId")]
                idName = "emdbId"
            self._setConstraintParameterMap(tableId=tableId, mapL=cMapL)  # pylint: disable=possibly-used-before-assignment
            #
            iCount = 0
            ifh = open(filePath, "r")
            accList = ifh.readlines()
            ifh.close()
            for acc in accList:
                if acc.startswith("#") or len(acc) < 3:
                    continue
                options = {}
                options[idName] = acc[:-1]
                nL = self._select(tableId=tableId, **options)
                if len(nL) > 0:
                    self.__lfh.write("WFTaskRequest.loadAccessions() skipping existing code %s\n" % acc[:-1])
                else:
                    options = {}
                    options[idName] = acc[:-1]
                    options["used"] = "n"
                    self._insertRequest(tableId=tableId, contextId=None, **options)
                    iCount += 1
            self.__lfh.write("WFTaskRequest.loadAccessions() loaded %d codes\n" % iCount)
        except:  # noqa: E722 pylint: disable=bare-except
            self.__lfh.write("WFTaskRequest.loadAccessions() failed \n")
            traceback.print_exc(file=self.__lfh)
            return False


if __name__ == "__main__":
    wfr = WFTaskRequest(None, verbose=True, log=sys.stderr)



##
# File:    taskReference.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#
##

"""
task object
"""

__docformat__ = "restructuredtext en"
__author__ = "Tom Oldfield"
__email__ = "oldfield@ebi.ac.uk"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.24"

import sys
from wwpdb.apps.wf_engine.reader.TaskComparitor import TaskComparitor
from wwpdb.apps.wf_engine.reader.TaskParameter import TaskParameter
from wwpdb.apps.wf_engine.reader.parseTimeDelta import parseTimeDelta

from wwpdb.utils.wf.process.ActionRegistry import ActionRegistry


class taskModule:

    """
    Holds the task object, the task information is passed
      out of the XML dom object
    """

    def __init__(self, debug=0, prt=sys.stderr):

        # string, double, complex, etc
        self.type = "unknown"
        # name of task for crossReference
        self.name = "unknown"
        # name of task for crossReference - human readable form
        self.nameHumanReadable = "unknown"
        # can we stop here for data
        self.breakPoint = "false"
        # exception hander : the task to handle expection
        self.exception = None
        # Used to store application name and file for workflows etc
        self.file = "unknown"
        # desciption - information for humans
        self.description = "unknown"
        # Not used   JDW
        # inputName - where the task flow was passed from
        # self.inputName = []
        #
        # outputName - where the task flow goes to.
        self.outputName = []
        # dataObject - list of data references
        self.data = []
        # expected run time
        self.runTime = 1
        # failsafe die
        self.failTime = 4
        # debug level
        self.debug = debug
        self.__lfh = prt
        # functions : list of functions
        self.func = []
        # parameters : 1 parameter
        self.param = None

        self.ActionRegistry = ActionRegistry()

        self.uniqueName = None
        self.uniqueType = None
        self.uniqueWhere = None
        self.uniqueAction = None
        self.uniqueActionName = None
        self.parameter = None
        #    self.uniqueContainer = None : put into data object
        #    self.uniqueType = None : put in data object

    def parseXML(self, taskNode):
        """parse the XML dom object to extract a task object"""

        self.name = taskNode.getAttribute("taskID")
        self.nameHumanReadable = taskNode.getAttribute("name")
        self.breakPoint = taskNode.getAttribute("breakpoint")
        self.exception = taskNode.getAttribute("exceptionID")
        self.uniqueActionName = taskNode.getAttribute("reference")
        if taskNode.nodeName == "wf:exception":
            self.outputName.append(taskNode.getAttribute("nextTask"))
            self.type = "exception"

        for taskType in taskNode.childNodes:
            if taskType.nodeName == "wf:description":
                self.description = taskType.firstChild.data
            elif taskType.nodeName == "wf:process":
                self.outputName.append(taskNode.getAttribute("nextTask"))
                self.parseProcessTask(taskType)
            elif taskType.nodeName == "wf:decision":
                self.parseAutoDecisionTask(taskType)
            elif taskType.nodeName == "wf:loop":
                self.outputName.append(taskNode.getAttribute("nextTask"))
                self.parseLoopTask(taskType)
            elif taskType.nodeName == "wf:workflow":
                self.type = "Workflow"
                self.file = taskType.getAttribute("file")
                self.outputName.append(taskNode.getAttribute("nextTask"))
                temp = taskType.getAttribute("runTime")
                self.runTime = int(parseTimeDelta(temp).total_seconds())
                temp = taskType.getAttribute("failTime")
                self.failTime = int(parseTimeDelta(temp).total_seconds())
            elif taskType.nodeName == "wf:handler":
                self.parseException(taskType)
            elif taskType.nodeName == "wf:manual":
                self.parseManualTask(taskType)
            elif taskType.nodeName == "wf:join":
                # this is not done at the moment - implied OR
                pass

    def parseException(self, exception):

        if self.debug > 0:
            pass
        datum = []
        # insert the default nextTask as the default handle
        self.outputName.append(exception.getAttribute("nextTask"))
        datum.append(exception.getAttribute("warning"))
        datum.append(exception.getAttribute("select"))
        datum.append(exception.getAttribute("nextTask"))
        datum.append("dummy1")
        datum.append("dummy2")
        self.data.append(datum)

        #    self.__lfh.write("+taskReference.parseException : data object (warning=<%40s> , select=<%15s>, nextTask = <%5s>\n" % (datum[0],datum[1],datum[2]));

    def parseLoopTask(self, loop):

        # for "value" in "iterator"
        self.type = "Loop"
        # no paramter - return value of list item,index : return index order of parameter ! ahhhh
        self.param = loop.getAttribute("parameter")
        # set the interator as the input to the loop
        datum = []
        datum.append(loop.getAttribute("iterator"))
        datum.append("input")
        self.data.append(datum)
        # set the value as the output to the loop
        datum = []
        datum.append(loop.getAttribute("value"))
        datum.append("output")
        self.data.append(datum)
        # there are 2 output tasks - the first is the loop next task
        # the second is the loop exit task
        self.outputName.append(loop.getAttribute("exitTask"))

    def parseAutoDecisionTask(self, decision):

        #   lookup = {"gte":">=","gt":">","lte":"<=","lt":"<","eq":"==","neq":"!=", "less":"<"}

        self.type = "Decision"
        self.file = decision.getAttribute("application")
        for info in decision.childNodes:
            if info.nodeName == "wf:dataObjectsLocation":
                for location in info.childNodes:
                    datum = []
                    if location.nodeName == "wf:location":
                        datum.append(str(location.getAttribute("dataID")))
                        datum.append(str(location.getAttribute("type")))
                        self.data.append(datum)
            elif info.nodeName == "wf:nextTasks":
                for output in info.childNodes:
                    if output.nodeName == "wf:nextTask":
                        self.outputName.append(output.getAttribute("taskID"))
                        for method in output.childNodes:
                            if method.nodeName == "wf:function":
                                #
                                compareAs = str(method.getAttribute("compareAs")).lower()
                                if compareAs not in ["int", "integer", "float", "double", "bool", "boolean", "string", "substring"]:
                                    # if omitted then apply a reasonable default  -
                                    if (method.getAttribute("string") != "") or (method.getAttribute("inList") != ""):
                                        compareAs = "string"
                                    else:
                                        compareAs = "integer"
                                f = TaskComparitor(self.debug, self.__lfh)
                                f.addDataName(method.getAttribute("dataID"))
                                if method.getAttribute("gte") != "":
                                    f.addData(-1, method.getAttribute("gte"), ">=", compareAs)
                                if method.getAttribute("gt") != "":
                                    f.addData(-1, method.getAttribute("gt"), ">", compareAs)
                                if method.getAttribute("lte") != "":
                                    f.addData(1, method.getAttribute("lte"), "<=", compareAs)
                                if method.getAttribute("less") != "":
                                    f.addData(1, method.getAttribute("less"), "<", compareAs)
                                if method.getAttribute("lt") != "":
                                    f.addData(1, method.getAttribute("lt"), "<", compareAs)
                                if method.getAttribute("eq") != "":
                                    f.addData(1, method.getAttribute("eq"), "==", compareAs)
                                if method.getAttribute("neq") != "":
                                    f.addData(1, method.getAttribute("neq"), "!=", compareAs)

                                if method.getAttribute("boolean") != "":
                                    f.addData(88, method.getAttribute("boolean"), compareAs="boolean")

                                if method.getAttribute("string") != "" and compareAs in ["string"]:
                                    f.addData(77, method.getAttribute("string"), compareAs=compareAs)
                                elif method.getAttribute("string") != "" and compareAs in ["substring"]:
                                    f.addData(78, method.getAttribute("string"), compareAs=compareAs)

                                elif method.getAttribute("inList") != "" and compareAs in ["string"]:
                                    f.addData(55, method.getAttribute("inList"), compareAs=compareAs)
                                elif method.getAttribute("inList") != "" and compareAs in ["substring"]:
                                    f.addData(56, method.getAttribute("inList"), compareAs=compareAs)
                        self.func.append(f)

    def parseProcessTask(self, process):

        self.type = "Process"
        temp = process.getAttribute("runTime")
        self.runTime = int(parseTimeDelta(temp).total_seconds())
        temp = process.getAttribute("failTime")
        self.failTime = int(parseTimeDelta(temp).total_seconds())

        for info in process.childNodes:
            if info.nodeName == "wf:detail":
                self.uniqueName = info.getAttribute("name")
                self.uniqueAction = info.getAttribute("action")
                if self.uniqueActionName is None or len(self.uniqueActionName) < 1:
                    self.uniqueActionName = self.uniqueAction
                self.parameter = info.getAttribute("parameter")
                self.uniqueWhere = info.getAttribute("where")
                if self.uniqueAction is not None:
                    if self.uniqueWhere == "api":
                        if not self.ActionRegistry.isDefinedAction(self.uniqueAction):
                            self.__lfh.write("+taskReference.parseException : **** Catastrophic WF error\n")
                            self.__lfh.write(
                                "+taskReference.parseProcessTask : Task refers to a APIprocess that does not exist : "
                                + str(self.uniqueAction)
                                + ", "
                                + str(self.uniqueWhere)
                                + "\n"
                            )
                            exit(0)
            #        self.uniqueContainer = info.getAttribute("returnContainer")
            #        self.uniqueType = info.getAttribute("returnType")
            elif info.nodeName == "wf:dataObjectsLocation":
                for data in info.childNodes:
                    datum = []
                    if data.nodeName == "wf:location":
                        datum.append(data.getAttribute("dataID"))
                        datum.append(data.getAttribute("type"))
                        self.data.append(datum)

    def parseManualTask(self, manual):

        self.type = "Manual"
        self.file = manual.getAttribute("application")
        for info in manual.childNodes:
            if info.nodeName == "wf:parameter":
                #        self.paramName = detail.getAttribute("name")
                n = 0
                self.param = TaskParameter()
                for write in info.childNodes:
                    if write.nodeName == "wf:write":
                        self.param.addTitle(write.getAttribute("title"))
                        self.param.addQuestion(write.getAttribute("question"))
                        for detail in write.childNodes:
                            if detail.nodeName == "wf:comment":
                                self.param.addComment(detail.firstChild.data)
                            elif detail.nodeName == "wf:objects":
                                for ob in detail.childNodes:
                                    if ob.nodeName == "wf:object":
                                        self.param.addObject(ob.getAttribute("ID"))
                                        self.param.addFormat(ob.getAttribute("format"))
            elif info.nodeName == "wf:nextTasks":
                for data in info.childNodes:
                    if data.nodeName == "wf:nextTask":
                        self.outputName.append(data.getAttribute("taskID"))
                        self.param.addOption(n, data.getAttribute("label"))
                        n = n + 1
            elif info.nodeName == "wf:dataObjectsLocation":
                for data in info.childNodes:
                    datum = []
                    if data.nodeName == "wf:location":
                        datum.append(data.getAttribute("dataID"))
                        datum.append(data.getAttribute("type"))
                        self.data.append(datum)

    def failTime(self):  # pylint: disable=method-hidden
        return self.failTime

    def runTime(self):  # pylint: disable=method-hidden
        return self.runTime

    def printName(self):
        """Convienence method to print the name of a task"""

        self.__lfh.write("+taskReference.printName : " + str(self.name) + "\n")
        return

    def printMe(self):
        """Convienence method to print things"""
        self.dump(self.__lfh)

    def dump(self, lfh):
        lfh.write("+taskReference.dump : Task type               %s\n" % self.type)
        lfh.write("+taskReference.dump : Task ID                 %s\n" % self.name)
        lfh.write("+taskReference.dump : Task name               %s\n" % self.nameHumanReadable)
        lfh.write("+taskReference.dump : Task breakpoint         %r\n" % self.breakPoint)
        lfh.write("+taskReference.dump : Task exception          %r\n" % self.exception)
        lfh.write("+taskReference.dump : Task file               %r\n" % self.file)
        lfh.write("+taskReference.dump : Task description        %s\n" % self.description)
        lfh.write("+taskReference.dump : Task unique name        %s\n" % self.uniqueName)
        lfh.write("+taskReference.dump : Task unique type        %s\n" % self.uniqueType)
        lfh.write("+taskReference.dump : Task unique where       %s\n" % self.uniqueWhere)
        lfh.write("+taskReference.dump : Task unique action      %s\n" % self.uniqueAction)
        lfh.write("+taskReference.dump : Task unique action name %s\n" % self.uniqueActionName)
        lfh.write("+taskReference.dump : Task output task names  %r\n" % self.outputName)

    def getDataReference(self, dataObjects):
        """Return the list of data objects from the input list related to the current task."""

        dat = []

        if self.data:
            for taskData in self.data:
                for dataRef in dataObjects:
                    if dataRef.isThisName(taskData[0]):
                        dataRef.localMode = taskData[1]
                        dat.append(dataRef)

        return dat

    def getName(self):
        return self.name

    def getFunc(self):
        return self.func

    def getNextTaskName(self, option=0):
        """Get the next task in the list"""

        whichOutput = 0
        for nextTask in self.outputName:
            if whichOutput == option:
                if self.debug > 0:
                    self.__lfh.write("+taskReference.getNextTaskName : next task " + str(nextTask) + " option " + str(option))
                return nextTask

            whichOutput = whichOutput + 1

        self.__lfh.write("+taskReference.getNextTaskName : Error : failed to find the next task by name\n")
        return None



##
# File:     SwitchTask.py
#
# Original:    autoDecideTask.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#
#   3-May-2015  jdw rewritten
#
##
"""

"""

import logging
from wwpdb.apps.wf_engine.engine.InterpretDataObject import getObjectValue
from wwpdb.apps.wf_engine.wf_engine_utils.run.MyLogger import MyLogger

logger = logging.getLogger(name="root")


class SwitchTask(object):

    """
    Handle switch/case comparison and branching workflow task
    """

    def __init__(self, task, debug=0):

        self.task = task
        self.debug = debug
        self.__lfh = MyLogger(level=logging.INFO)

    def handleTask(self, valueD):
        if self.task is None:
            return -1
        logger.info("+SwitchTask.getFunctionReturn : starting for task %s valueD %r\n", self.task.name, valueD)
        funcList = self.task.getFunc()
        count = 0
        sValue = None
        for f in funcList:
            if f is None or f.getDataName() not in valueD:
                return -1
            data = valueD[f.getDataName()]
            sValue = getObjectValue(data, self.debug, self.__lfh)

            if sValue is None:
                return -1

            if self.debug > 0:
                logger.info("+SwitchTask.getFunctionReturn :  test value  %r", sValue)
                f.printMe(self.__lfh)

            if f.check(sValue) == 1:
                return count
            count = count + 1
        #
        logger.info("+SwitchTask.getFunctionReturn :  comparison failed with sValue %r\n ", sValue)
        return -1



##
# File:    ExternalTask.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#  #
#  15-Feb-2015   jdw - Gutted -  strip out broken gui console
#   5-Apr-2015   jdw   Adjust logging output -
#  2-May-2015    jdw   refactor -
#  1-Jun-2016    jdw   review for restart/recovery issues -
#
##

"""
The external task processor.
"""


import time
import sys
import datetime
import logging

from wwpdb.apps.wf_engine.wf_engine_utils.time.TimeStamp import TimeStamp

logger = logging.getLogger(name="root")


class ExternalTask(object):

    """
    This task manager is for the control and access to the manual interfaces

    All communication is via the statusDB
      handleExternalTask : puts in "waiting"
      This method is aware of an interface "open" ; but takes no action (for now)
      The exit condition is "closed(<N>)" within the statusDB; N = return condition from interface
    Note :
      The WFM uses the instance data for testing, the WFE is based on task data.  Therefore this
        method uses both the instance and task data for communication.

    """

    def __init__(self, eUtil, depositionID, WorkflowClassID, WorkflowInstanceID, task, debug=0, recovery=0):

        self.task = task
        self.__debug = debug
        self.WorkflowInstanceID = WorkflowInstanceID
        self.WorkflowClassID = WorkflowClassID
        self.depositionID = depositionID
        self.recovery = recovery
        self.__timeStamp = TimeStamp()
        self.__eUtil = eUtil
        logger.info("\n+ExternalTask.__init__ : ------------------------------------------------------------------------------------")
        logger.info(
            "+ExternalTask.__init__ : depositionId %r workflowclassid %r workflowinstanceid %r task.name %r recovery %r\n",
            depositionID,
            WorkflowClassID,
            WorkflowInstanceID,
            task.name,
            recovery,
        )

    def __setDBTaskStatus(self, typein, mode):  # pylint: disable=unused-argument
        """
        Method to update the wf_task table - audit trail data for WFE

        type: argument not used --
        """

        taskID = {}
        taskID["WF_TASK_ID"] = self.task.name
        taskID["WF_INST_ID"] = self.WorkflowInstanceID
        taskID["WF_CLASS_ID"] = self.WorkflowClassID
        taskID["DEP_SET_ID"] = self.depositionID
        taskID["STATUS_TIMESTAMP"] = self.__timeStamp.getSecondsFromReference()

        self.__eUtil.updateStatus(taskID, mode)

    def __setDBInstStatus(self, typein, mode):  # pylint: disable=unused-argument
        """
        Method to update the wf_instance table - WFM communication

        type: argument not used -

        """

        now = self.__timeStamp.getSecondsFromReference()
        instDB = {}
        instDB["WF_INST_ID"] = self.WorkflowInstanceID
        instDB["WF_CLASS_ID"] = self.WorkflowClassID
        instDB["DEP_SET_ID"] = self.depositionID
        instDB["STATUS_TIMESTAMP"] = now
        self.__eUtil.updateStatus(instDB, mode)
        # Tom :  update the wf_instance_last table - note it is unique over dep_set_id

        sql = (
            "update wf_instance_last set status_timestamp="
            + str(now)
            + ", inst_status='"
            + mode
            + "', wf_class_id = '"
            + self.WorkflowClassID
            + "', wf_inst_id = '"
            + self.WorkflowInstanceID
            + "' where dep_set_id = '"
            + self.depositionID
            + "'"
        )
        ok = self.__eUtil.runUpdateSQL(sql)
        if ok < 1:
            logger.info("+ExternalTask.__setDBInstStatus - CRITICAL : failed to update wf_instance_last table\n")

    def __getDBInstStatus(self):
        """
        Method to get the status of the instance - for WFM communication
        """

        self.__eUtil.updateConnection()

        instID = self.__eUtil.getObject(self.depositionID, self.WorkflowClassID, self.WorkflowInstanceID)

        if instID is None:
            logger.info("+ExternalTask.__getDBInstStatus :  *** Database went away ***\n")
            logger.info("+ExternalTask.__getDBInstStatus : Exception : database went away\n")
            logger.info("+ExternalTask.__getDBInstStatus : Manual task : getDBInstStatus \n")
            logger.info("+ExternalTask.__getDBInstStatus : UTC time = %s\n", str(datetime.datetime.utcnow()))
            sys.exit(0)

        else:
            ret = self.__eUtil.getStatus(instID)

        return ret

    def handleTask(self):
        """
        The main wait function for the manual interface.  loop for wait
        for the statusDB to change.

        Note that the wait is on the instance state (to reduce contention for  WFM and interfaces)

        """
        logger.info(
            "\n+ExternalTask.handleExternalTask : starts with id %s  class %s instance %s task.name %r recovery %r",
            self.depositionID,
            self.WorkflowClassID,
            self.WorkflowInstanceID,
            self.task.name,
            self.recovery,
        )
        # set the instance and task  status to waiting
        if self.recovery == 0:
            logger.info("+ExternalTask.handleExternalTask : depositionID %r task.name %r set task status WAITING", self.depositionID, self.task.name)
            self.__setDBTaskStatus("manual", "waiting")

        # put in a passthrough for restarting interfaces - ie if the UI has put closed into
        # the status DB then we don't want to overwrite this - just pass stright through and capture this
        if self.recovery == 0:
            logger.info("+ExternalTask.handleExternalTask : depositionID %r task.name %r set instance status WAITING", self.depositionID, self.task.name)
            self.__setDBInstStatus("manual", "waiting")

        # A short sleep to make sure everyone is up and waiting.
        try:
            time.sleep(1.0)
        except Exception as _e:  # noqa: F841
            logger.info("+ExternalTask.handleExternalTask :  Exception during timer %s ", self.depositionID)
            return str(-1)

        startTime = datetime.datetime.now()

        opened = False
        while True:
            state = self.__getDBInstStatus()

            if self.__debug > 3:
                elapsedT = datetime.datetime.now() - startTime
                logger.debug(
                    "+ExternalTask.handleExternalTask :  id %s  class %s instance %s elapsed (secs) %d state %r ",
                    self.depositionID,
                    self.WorkflowClassID,
                    self.WorkflowInstanceID,
                    elapsedT.total_seconds(),
                    str(state),
                )

            if state[:6] == "closed":
                n1 = state.find("(")
                n2 = state.find(")")
                if n1 >= 0 and n2 > (n1 + 1):
                    ret = state[n1 + 1 : n2]
                else:
                    # we get the default return state from the interface
                    ret = "0"
                # reset the instance state for the WFM
                self.__setDBInstStatus("manual", "running")
                logger.info("+ExternalTask.handleExternalTask : depositionID %r detected CLOSED instance status - reset to running", self.depositionID)
                break
            elif state == "close(o)" or state == "close(0)" or state == "closed()":
                # cludge from wrong interface return - probably don't need this
                ret = "0"
                # reset the instance state for the WFM
                self.__setDBInstStatus("manual", "running")
                logger.info("+ExternalTask.handleExternalTask : depositionID %r detected CLOSED instance status - set to running", self.depositionID)
                break
            elif state == "open":
                # still waiting - but the interface is now open
                if not opened:
                    logger.info("+ExternalTask.handleExternalTask : depositionID %r detected OPEN instance status - update LAST instance", self.depositionID)
                    opened = True
                    # manage the UI not knowing about this new table
                    sql = (
                        "update wf_instance_last set status_timestamp="
                        + str(self.__timeStamp.getSecondsFromReference())
                        + ", inst_status='open', wf_class_id = '"
                        + self.WorkflowClassID
                        + "', wf_inst_id = '"
                        + self.WorkflowInstanceID
                        + "' where dep_set_id = '"
                        + self.depositionID
                        + "'"
                    )

                    _ok = self.__eUtil.runUpdateSQL(sql)  # noqa: F841
            elif state == "waiting":
                # still waiting - no one has opened the interface
                pass  # pylint: disable=unnecessary-pass
            elif state == "exception":
                # the interface threw and exception code
                logger.info("+ExternalTask.handleExternalTask : depositionID %r detected EXCEPTION instance status - reset to exception", self.depositionID)
                self.__setDBInstStatus("manual", "exception")
                ret = str(-1)
                break
            elif state == "aborted":
                # the interface threw and exception code
                logger.info("+ExternalTask.handleExternalTask : depositionID %r detected ABORTED instance status - reset to ABORTED", self.depositionID)
                self.__setDBInstStatus("manual", "aborted")
                ret = str(-1)
                break

            try:
                time.sleep(0.7)
            except Exception as _e:  # noqa: F841
                # someone killed the WFE wait : register an exception
                logger.info("+ExternalTask.handleExternalTask :  Exception during timer %s ", self.depositionID)
                self.__setDBInstStatus("manual", "exception")
                ret = str(-1)
                break

        logger.info(
            "\n+ExternalTask.handleExternalTask : ENDS for id %s  class %s instance %s task.name %r recovery %r return value %r",
            self.depositionID,
            self.WorkflowClassID,
            self.WorkflowInstanceID,
            self.task.name,
            self.recovery,
            ret,
        )
        #
        return int(ret)



#
# Regular expression to parse a timedelta
#
# Taken from https://stackoverflow.com/questions/12075562/how-to-deal-with-time-values-over-24-hours-in-python
# Updated http://kbyanc.blogspot.com/2007/08/python-reconstructing-timedeltas-from.html
#
#
import re
from datetime import timedelta


def parseTimeDelta(s):
    """Create timedelta object representing time delta
       expressed in a string

    Takes a string in the format produced by calling str() on
    a python timedelta object and returns a timedelta instancee
    that would produce that string.

    Acceptable formats are: "X days, HH:MM:SS" or "HH:MM:SS".
    """
    if s is None:
        return None
    d = re.match(r"((?P<days>\d+) days, )?(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)", str(s)).groupdict(0)
    return timedelta(**dict(((key, int(value)) for key, value in d.items())))


if __name__ == "__main__":
    print(parseTimeDelta("24:00:00"))
    print(parseTimeDelta("24:00:00").total_seconds())
    print(parseTimeDelta("28:32:12"))



##
# File:    dataReference.py
# Date:    15-Mar-2009
#
# Updates:
#  22-April-2010 : Incorporation into enterpise.
#
##

#
# data object definition
#

__docformat__ = "restructuredtext en"
__author__ = "Tom Oldfield"
__email__ = "oldfield@ebi.ac.uk"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.24"

import sys

from wwpdb.utils.wf.WfDataObject import WfDataObject


class dataModule(object):

    """Hold a data reference, passed variable is the XML dom
    object to be parsed"""

    def __init__(self, instanceID, depositionID, debug=0, prt=sys.stderr):

        self.debug = debug
        self.__lfh = prt

        # string, double, complex etc...
        self.type = "unknown"
        # reference name for lookup from task
        self.name = "unknown"
        # reference name for lookup from task - humanReadableForm
        self.nameHumanReadable = "unknown"
        # description is just a string for humans
        self.description = "unknown"
        # could be File, DataModel, WorkflowData
        self.where = "unknown"
        # localMode is the task mode for that data instance
        self.localMode = "unknown"
        # mutability of data - is it write-locked
        self.mutable = "true"
        # and the data for the WFE data only
        self.data = None
        # depositionID
        # create a apiDataobject
        self.ApiData = WfDataObject()
        self.ApiData.setDepositionDataSetId(depositionID)
        if debug > 1:
            self.__lfh.write("+dataModule.__init__ : depositionID " + str(depositionID) + "\n")

        self.ApiData.setWorkflowInstanceId(instanceID)
        if debug > 1:
            self.__lfh.write("+dataModule.__init__ :  instanceID  " + str(instanceID) + "\n")
        # status code for the data reading
        self.status = "OK"

        # containers for all API attributes
        self.valueContainer = None
        self.valueType = None
        self.version = None
        self.value = None
        self.format = None
        self.content = None
        self.selectCategory = None
        self.selectAttribute = None
        self.selectConditionAttribute = None
        self.partitionNumber = None
        self.referenceType = None

    def getValue(self):
        return self.data

    def setValue(self, value):
        self.data = value

    def setInstanceID(self, instanceID):

        self.ApiData.setWorkflowInstanceId(instanceID)
        if self.debug > 1:
            self.__lfh.write("+dataModule.setInstanceId:  setWorkflowInstanceId  " + str(instanceID) + "\n")

    def parseXML(self, data):
        #   class = dataModule : fill class variable from dom

        self.type = data.getAttribute("type")
        self.name = data.getAttribute("dataID")
        self.nameHumanReadable = data.getAttribute("name")

        self.valueContainer = data.getAttribute("container")
        self.valueType = data.getAttribute("type")
        self.mutable = data.getAttribute("mutable")

        for detail in data.childNodes:
            if detail.nodeName == "wf:description":
                self.description = detail.firstChild.data
            elif detail.nodeName == "wf:location":
                self.where = str(detail.getAttribute("where"))
                # if self.where == "archive" or self.where == "wf-instance" or self.where[:4] == "path" or self.where == 'deposit':
                if (len(self.where) > 0) and (self.where not in ["constant", "inline"]):
                    self.version = detail.getAttribute("version")
                    self.content = detail.getAttribute("content")
                    self.format = detail.getAttribute("format")
                    self.selectCategory = str(detail.getAttribute("selectCategory"))
                    self.selectAttribute = str(detail.getAttribute("selectAttribute"))
                    self.partitionNumber = str(detail.getAttribute("partitionNumber"))
                    self.selectConditionAttribute = str(detail.getAttribute("selectConditionAttribute"))
                    self.referenceType = str(detail.getAttribute("referenceType"))

                # WF data objects that are initialized values within workflow definitions.
                elif self.where == "constant":
                    if self.debug > 1:
                        self.__lfh.write("+dataModule.parseXML : found constant data object\n")
                        self.__lfh.write("+dataModule.parseXML : found container " + str(self.valueContainer) + "\n")
                        self.__lfh.write("+dataModule.parseXML : found type " + str(self.valueType) + "\n")
                        self.__lfh.write("+dataModule.parseXML : found value " + str(detail.getAttribute("value")) + "\n")
                    if self.valueContainer == "list":
                        s = detail.getAttribute("value")
                        self.data = []
                        self.data = s.split(",")
                    elif self.valueContainer == "string":
                        self.data = detail.getAttribute("value")
                    if self.debug > 1:
                        self.__lfh.write("+dataModule.parseXML : value = " + str(self.data))
                elif self.where == "inline":
                    if self.debug > 1:
                        self.__lfh.write("+dataModule.parseXML : found inline data object\n")
                        self.__lfh.write("+dataModule.parseXML : found container " + str(self.valueContainer) + "\n")
                        self.__lfh.write("+dataModule.parseXML : found type " + str(self.valueType) + "\n")
                        self.__lfh.write("+dataModule.parseXML : found value " + str(detail.getAttribute("value")) + "\n")
                    if self.valueContainer == "list":
                        self.data = []
                    elif self.valueContainer == "string":
                        self.data = ""
                    if self.debug > 1:
                        self.__lfh.write("+dataModule.parseXML : inline variable initialized with = " + str(self.data))

        if self.debug > 1:
            self.printMe(self.__lfh)

    def setLocalMode(self, local):
        # set the value of the local mode for Class = dataModule

        self.localMode = local

    def isThisName(self, compareName):

        if self.name == compareName:
            return "yes"
        else:
            return ""

    def printMe(self, lfh):
        lfh.write("+dataModule.printMe : Data ID               %s\n" % self.name)
        lfh.write("+dataModule.printMe : Data name             %s\n" % self.nameHumanReadable)
        lfh.write("+dataModule.printMe : Data description      %s\n" % self.description)
        lfh.write("+dataModule.printMe : Data Type             %s\n" % self.type)
        lfh.write("+dataModule.printMe : Data location(where)  %s\n" % self.where)
        lfh.write("+dataModule.printMe : Data Container        %s\n" % self.valueContainer)
        lfh.write("+dataModule.printMe : Data type             %s\n" % self.valueType)
        if self.data is not None:
            lfh.write("+dataModule.printMe : Data object value     %r\n" % self.data)
        if self.localMode:
            lfh.write("+dataModule.printMe : Data Mode             %s\n" % self.localMode)
        #



##
# File:    TaskComparitor.py
#
# Updates:
#
#  3-May-2015   jdw rewritten  -
#
##

"""
TaskComparitor() handles comparisons for SwitchTask()
"""

import sys


class TaskComparitor(object):
    def __init__(self, debug=0, prt=sys.stderr):  # pylint: disable=unused-argument

        self.__dataId = None
        self.min = 0
        self.max = 0
        self.lower = ">"
        self.upper = "<"
        self.logical = False
        self.string = None
        self.__compareCode = None
        self.__compareAs = None

        self.debug = debug
        # self.__lfh = prt

    def addDataName(self, name):
        self.__dataId = name

    def getDataName(self):
        return self.__dataId

    def __applyType(self, compareAs, sValue):
        if compareAs not in ["int", "integer"]:
            lData = int(sValue)
        elif compareAs in ["float", "double"]:
            lData = float(sValue)
        elif compareAs in ["string", "substring"]:
            lData = str(sValue)
        else:
            lData = int(sValue)
        return lData

    def addData(self, compareCode, data, mode="<", compareAs="integer"):
        self.__compareCode = compareCode
        self.__compareAs = compareAs
        if compareAs in ["int", "integer"]:
            lData = int(data)
        elif compareAs in ["float", "double"]:
            lData = float(data)
        elif compareAs in ["string", "substring"]:
            lData = str(data)
        else:
            lData = int(data)

        if compareCode == -1:
            self.min = lData
            self.lower = mode
        elif compareCode == 1:
            self.max = lData
            self.upper = mode
        elif compareCode == 0:
            self.min = lData
            self.max = lData
            self.lower = mode
            self.upper = mode
        elif compareCode == 88:
            if data.lower() == "true" or data.lower() == "ok":
                self.logical = False
            else:
                self.logical = True
        elif compareCode in [77, 78, 55, 56]:
            self.string = str(data)

    def check(self, value):
        if self.__compareCode in [-1, 1, 0]:
            ret = self.__checkValue(value)
        elif self.__compareCode in [88]:
            ret = self.__checkBoolean(value)
        elif self.__compareCode in [77]:
            ret = self.__checkString(value)
        elif self.__compareCode in [78]:
            ret = self.__checkSubString(value)  # XX Not implemented yet  pylint: disable=no-member
        elif self.__compareCode in [55]:
            ret = self.__checkStringInList(value)
        elif self.__compareCode in [56]:
            ret = self.__checkSubStringInList(value)
        else:
            ret = 0
        return ret

    def printMe(self, lfh):
        lfh.write("+taskFunction.printMe()  min %s max %s lower %s upper %s\n" % (self.min, self.max, self.lower, self.upper))
        lfh.write(
            "+taskFunction.printMe()  dataId %s compare code %r compareAs %s logical %r string %s\n"
            % (self.__dataId, self.__compareCode, self.__compareAs, self.logical, self.string)
        )

    def __checkBoolean(self, value):
        if value:
            if self.logical:
                return 1
            else:
                return 0
        else:
            if self.logical:
                return 0
            else:
                return 1

    def __checkString(self, value):

        if value is None:
            if self.string.lower() == "none":
                return 1
            else:
                return 0

        if str(value).lower() == self.string.lower():
            return 1
        else:
            return 0

    def __checkStringInList(self, valList):

        if valList is None:
            return 0
        if not isinstance(valList, list):
            return 0
        try:
            for val in valList:
                if str(self.string).lower() == str(val).lower():
                    return 1
            return 0
        except:  # noqa: E722 pylint: disable=bare-except
            return 0

    def __checkSubStringInList(self, valList):
        if valList is None:
            return 0
        if not isinstance(valList, list):
            return 0
        try:
            for val in valList:
                if str(self.string).lower() in str(val).lower():
                    return 1
            return 0
        except:  # noqa: E722 pylint: disable=bare-except
            return 0

    def __checkValue(self, sValue):

        value = self.__applyType(self.__compareAs, sValue)

        if self.lower == ">" and self.upper == "<":
            if value > self.min and value < self.max:
                return 1
            else:
                return 0
        elif self.lower == ">=" and self.upper == "<":
            if value >= self.min and value < self.max:
                return 1
            else:
                return 0
        elif self.lower == ">" and self.upper == "<=":
            if value > self.min and value <= self.max:
                return 1
            else:
                return 0
        elif self.lower == ">=" and self.upper == "<=":
            if value >= self.min and value <= self.max:
                return 1
            else:
                return 0
        elif self.lower == "==" and self.upper == "==":
            if value == self.min and value == self.max:
                return 1
            else:
                return 0
        elif self.lower == "!=" and self.upper == "!=":
            if value == self.min and value == self.max:
                return 0
            else:
                return 1

        else:
            return 0



#!/usr/bin/env python
##
# File: WFEngineRunnerExec.py
# Date: 12-Feb-2015
#
# Updates:
#   5-Feb-2016 jdw provide host specific stdout/stderr logs --
#   4-Mar-2016 jdw add a more restrictive file permissions mask
##
"""
Set up execution envirnoment for workflow engine -

"""
__docformat__ = "restructuredtext en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Creative Commons Attribution 3.0 Unported"
__version__ = "V0.001"

import sys
import os
import time
import logging

from optparse import OptionParser  # pylint: disable=deprecated-module
from wwpdb.apps.wf_engine.wf_engine_utils.run.WFEngineRunner import WFEngineRunner

import platform

from wwpdb.utils.config.ConfigInfo import ConfigInfo, getSiteId


def main():
    # adding a conservative permission mask for this
    os.umask(0o022)
    #
    siteId = getSiteId(defaultSiteId="WWPDB_DEPLOY_TEST")
    cI = ConfigInfo(siteId)

    #    topPath = cI.get('SITE_WEB_APPS_TOP_PATH')
    topSessionPath = cI.get("SITE_WEB_APPS_TOP_SESSIONS_PATH")
    wfXmlPath = cI.get("SITE_WF_XML_PATH")
    #
    myFullHostName = platform.uname()[1]
    myHostName = str(myFullHostName.split(".")[0]).lower()
    #

    wfLogDirPath = os.path.join(topSessionPath, "wf-logs")
    if not os.path.exists(wfLogDirPath):
        os.makedirs(wfLogDirPath)
    pidFilePath = os.path.join(wfLogDirPath, myHostName + ".pid")
    #
    stdoutFilepath = os.path.join(wfLogDirPath, myHostName + "-stdout.log")
    stderrFilepath = os.path.join(wfLogDirPath, myHostName + "-stderr.log")
    #
    #  Setup logging  --
    now = time.strftime("-%Y-%m-%d", time.localtime())
    wfLogFilePath = os.path.join(wfLogDirPath, myHostName + now + ".log")
    logger = logging.getLogger(name="root")
    logging.captureWarnings(True)
    #
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(module)s - %(message)s")
    handler = logging.FileHandler(wfLogFilePath)
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    #

    usage = "usage: %prog [options]"
    parser = OptionParser(usage)
    parser.add_option("--start", default=False, action="store_true", dest="startOp", help="Start workflow engine process")
    parser.add_option("--stop", default=False, action="store_true", dest="stopOp", help="Stop workflow process")
    parser.add_option("--restart", default=False, action="store_true", dest="restartOp", help="Restart workflow engine process")
    parser.add_option("--status", default=False, action="store_true", dest="statusOp", help="Report workflow enging process status")

    parser.add_option("-v", "--verbose", default=False, action="store_true", dest="verbose", help="Enable verbose output")
    parser.add_option("--debug", default=1, type="int", dest="debugLevel", help="Debug level [0-4]")
    parser.add_option("--xmlpath", default=None, dest="wfXmlPath", help="Workflow XML definition file path")

    (options, _args) = parser.parse_args()
    lt = time.strftime("%Y %m %d %H:%M:%S", time.localtime())
    #
    if options.debugLevel > 2:
        logger.setLevel(logging.DEBUG)
    elif options.debugLevel > 0:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.ERROR)
    if options.wfXmlPath is not None:
        wfXmlPath = os.path.abspath(options.wfXmlPath)

    wfr = WFEngineRunner(pidFile=pidFilePath, wfXmlPath=wfXmlPath, debugLevel=options.debugLevel, stdout=stdoutFilepath, stderr=stderrFilepath, wrkDir=wfLogDirPath)

    if options.startOp:
        sys.stdout.write("+WFEngineRunnerExec(main) starting workflow engine at %s\n" % lt)
        logger.info("+WFEngineRunnerExec(main) starting workflow engine at %s\n", lt)
        wfr.setDebugLevel(level=options.debugLevel)
        wfr.start()
    elif options.stopOp:
        sys.stdout.write("+WFEngineRunnerExec(main) stopping workflow engine at %s\n" % lt)
        logger.info("+WFEngineRunnerExec(main) starting workflow engine at %s\n", lt)
        wfr.stop()
    elif options.restartOp:
        sys.stdout.write("+WFEngineRunnerExec(main) restarting workflow engine at %s\n" % lt)
        wfr.setDebugLevel(level=options.debugLevel)
        wfr.restart()
    elif options.statusOp:
        sys.stdout.write("+WFEngineRunnerExec(main) reporting workflow engine status at %s\n" % lt)
        sys.stdout.write(wfr.status())
    else:
        pass


if __name__ == "__main__":
    main()
